import pytest

import kopf


def test_adding_to_dict():
    obj = {}
    kopf.label(obj, {'label-1': 'value-1', 'label-2': 'value-2'})
    assert 'metadata' in obj
    assert 'labels' in obj['metadata']
    assert isinstance(obj['metadata']['labels'], dict)
    assert len(obj['metadata']['labels']) == 2
    assert 'label-1' in obj['metadata']['labels']
    assert 'label-2' in obj['metadata']['labels']
    assert obj['metadata']['labels']['label-1'] == 'value-1'
    assert obj['metadata']['labels']['label-2'] == 'value-2'


def test_adding_to_dicts(multicls):
    obj1 = {}
    obj2 = {}
    objs = multicls([obj1, obj2])
    kopf.label(objs, {'label-1': 'value-1', 'label-2': 'value-2'})
    assert isinstance(obj1['metadata']['labels'], dict)
    assert len(obj1['metadata']['labels']) == 2
    assert 'label-1' in obj1['metadata']['labels']
    assert 'label-2' in obj1['metadata']['labels']
    assert obj1['metadata']['labels']['label-1'] == 'value-1'
    assert obj1['metadata']['labels']['label-2'] == 'value-2'

    assert isinstance(obj2['metadata']['labels'], dict)
    assert len(obj2['metadata']['labels']) == 2
    assert 'label-1' in obj2['metadata']['labels']
    assert 'label-2' in obj2['metadata']['labels']
    assert obj2['metadata']['labels']['label-1'] == 'value-1'
    assert obj2['metadata']['labels']['label-2'] == 'value-2'


def test_adding_to_pykube_object(pykube_object):
    del pykube_object.obj['metadata']
    kopf.label(pykube_object, {'label-1': 'value-1', 'label-2': 'value-2'})
    assert len(pykube_object.labels) == 2
    assert 'label-1' in pykube_object.labels
    assert 'label-2' in pykube_object.labels
    assert pykube_object.labels['label-1'] == 'value-1'
    assert pykube_object.labels['label-2'] == 'value-2'


def test_adding_to_kubernetes_model(kubernetes_model):
    kubernetes_model.metadata = None
    kopf.label(kubernetes_model, {'label-1': 'value-1', 'label-2': 'value-2'})
    assert len(kubernetes_model.metadata.labels) == 2
    assert 'label-1' in kubernetes_model.metadata.labels
    assert 'label-2' in kubernetes_model.metadata.labels
    assert kubernetes_model.metadata.labels['label-1'] == 'value-1'
    assert kubernetes_model.metadata.labels['label-2'] == 'value-2'


def test_adding_to_kubernetes_asyncio_model(kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata = None
    kopf.label(kubernetes_asyncio_model, {'label-1': 'value-1', 'label-2': 'value-2'})
    assert len(kubernetes_asyncio_model.metadata.labels) == 2
    assert 'label-1' in kubernetes_asyncio_model.metadata.labels
    assert 'label-2' in kubernetes_asyncio_model.metadata.labels
    assert kubernetes_asyncio_model.metadata.labels['label-1'] == 'value-1'
    assert kubernetes_asyncio_model.metadata.labels['label-2'] == 'value-2'


def test_forcing_true_warns_on_deprecated_option():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    with pytest.deprecated_call(match=r"use forced="):
        kopf.label(obj, {'label': 'new-value'}, force=True)
    assert obj['metadata']['labels']['label'] == 'new-value'


def test_forcing_false_warns_on_deprecated_option():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    with pytest.deprecated_call(match=r"use forced="):
        kopf.label(obj, {'label': 'new-value'}, force=False)
    assert obj['metadata']['labels']['label'] == 'old-value'


def test_forcing_true_to_dict():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    kopf.label(obj, {'label': 'new-value'}, forced=True)
    assert obj['metadata']['labels']['label'] == 'new-value'


def test_forcing_false_to_dict():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    kopf.label(obj, {'label': 'new-value'}, forced=False)
    assert obj['metadata']['labels']['label'] == 'old-value'


def test_forcing_default_to_dict():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    kopf.label(obj, {'label': 'new-value'})
    assert obj['metadata']['labels']['label'] == 'old-value'


def test_forcing_true_to_pykube_object(pykube_object):
    pykube_object.labels['label'] = 'old-value'
    kopf.label(pykube_object, {'label': 'new-value'}, forced=True)
    assert pykube_object.labels['label'] == 'new-value'


def test_forcing_false_to_pykube_object(pykube_object):
    pykube_object.labels['label'] = 'old-value'
    kopf.label(pykube_object, {'label': 'new-value'}, forced=False)
    assert pykube_object.labels['label'] == 'old-value'


def test_forcing_default_to_pykube_object(pykube_object):
    pykube_object.labels['label'] = 'old-value'
    kopf.label(pykube_object, {'label': 'new-value'})
    assert pykube_object.labels['label'] == 'old-value'


def test_forcing_true_to_kubernetes_model(kubernetes_model):
    kubernetes_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_model, {'label': 'new-value'}, forced=True)
    assert kubernetes_model.metadata.labels['label'] == 'new-value'


def test_forcing_false_to_kubernetes_model(kubernetes_model):
    kubernetes_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_model, {'label': 'new-value'}, forced=False)
    assert kubernetes_model.metadata.labels['label'] == 'old-value'


def test_forcing_default_to_kubernetes_model(kubernetes_model):
    kubernetes_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_model, {'label': 'new-value'})
    assert kubernetes_model.metadata.labels['label'] == 'old-value'


def test_forcing_true_to_kubernetes_asyncio_model(kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_asyncio_model, {'label': 'new-value'}, forced=True)
    assert kubernetes_asyncio_model.metadata.labels['label'] == 'new-value'


def test_forcing_false_to_kubernetes_asyncio_model(kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_asyncio_model, {'label': 'new-value'}, forced=False)
    assert kubernetes_asyncio_model.metadata.labels['label'] == 'old-value'


def test_forcing_default_to_kubernetes_asyncio_model(kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_asyncio_model, {'label': 'new-value'})
    assert kubernetes_asyncio_model.metadata.labels['label'] == 'old-value'


@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_true_to_dict(nested):
    obj = {'metadata': {'labels': {'label': 'old-value'}},
           'spec': {'jobTemplate': {}}}
    kopf.label(obj, {'label': 'new-value'}, nested=nested, forced=True)
    assert obj['metadata']['labels']['label'] == 'new-value'
    assert obj['spec']['jobTemplate']['metadata']['labels']['label'] == 'new-value'
    assert 'unexistent' not in obj['spec']


@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_true_to_pykube_object(nested, pykube_object):
    pykube_object.labels.update({'label': 'old-value'})
    pykube_object.obj.update({'spec': {'jobTemplate': {}}})
    kopf.label(pykube_object, {'label': 'new-value'}, nested=nested, forced=True)
    assert pykube_object.labels['label'] == 'new-value'
    assert pykube_object.obj['spec']['jobTemplate']['metadata']['labels']['label'] == 'new-value'
    assert 'unexistent' not in pykube_object.obj['spec']


@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_true_to_kubernetes_model(nested, kubernetes_model):
    kubernetes_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_model, {'label': 'new-value'}, nested=nested, forced=True)
    assert kubernetes_model.metadata.labels['label'] == 'new-value'
    assert kubernetes_model.spec.job_template.metadata.labels['label'] == 'new-value'
    assert not hasattr(kubernetes_model.spec, 'unexistent')



@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_true_to_kubernetes_asyncio_model(nested, kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_asyncio_model, {'label': 'new-value'}, nested=nested, forced=True)
    assert kubernetes_asyncio_model.metadata.labels['label'] == 'new-value'
    assert kubernetes_asyncio_model.spec.job_template.metadata.labels['label'] == 'new-value'
    assert not hasattr(kubernetes_asyncio_model.spec, 'unexistent')


@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_false_to_dict(nested):
    obj = {'metadata': {'labels': {'label': 'old-value'}},
           'spec': {'jobTemplate': {}}}
    kopf.label(obj, {'label': 'new-value'}, nested=nested, forced=False)
    assert obj['metadata']['labels']['label'] == 'old-value'
    assert obj['spec']['jobTemplate']['metadata']['labels']['label'] == 'new-value'
    assert 'unexistent' not in obj['spec']


@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_false_to_pykube_object(nested, pykube_object):
    pykube_object.labels.update({'label': 'old-value'})
    pykube_object.obj.update({'spec': {'jobTemplate': {}}})
    kopf.label(pykube_object, {'label': 'new-value'}, nested=nested, forced=False)
    assert pykube_object.labels['label'] == 'old-value'
    assert pykube_object.obj['spec']['jobTemplate']['metadata']['labels']['label'] == 'new-value'
    assert 'unexistent' not in pykube_object.obj['spec']


@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_false_to_kubernetes_model(nested, kubernetes_model):
    kubernetes_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_model, {'label': 'new-value'}, nested=nested, forced=False)
    assert kubernetes_model.metadata.labels['label'] == 'old-value'
    assert kubernetes_model.spec.job_template.metadata.labels['label'] == 'new-value'
    assert not hasattr(kubernetes_model.spec, 'unexistent')


@pytest.mark.parametrize('nested', [
    pytest.param(('spec.jobTemplate', 'spec.unexistent'), id='tuple'),
    pytest.param(['spec.jobTemplate', 'spec.unexistent'], id='list'),
    pytest.param({'spec.jobTemplate', 'spec.unexistent'}, id='set'),
    pytest.param('spec.jobTemplate', id='string'),
])
def test_nested_with_forced_false_to_kubernetes_asyncio_model(nested, kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata.labels = {'label': 'old-value'}
    kopf.label(kubernetes_asyncio_model, {'label': 'new-value'}, nested=nested, forced=False)
    assert kubernetes_asyncio_model.metadata.labels['label'] == 'old-value'
    assert kubernetes_asyncio_model.spec.job_template.metadata.labels['label'] == 'new-value'
    assert not hasattr(kubernetes_asyncio_model.spec, 'unexistent')



from unittest.mock import Mock

import pytest

import kopf
from kopf._core.engines.admission import serve_admission_request


@pytest.mark.parametrize('dryrun', [True, False])
async def test_dryrun_passed(
        settings, registry, resource, memories, insights, indices, adm_request, dryrun):
    mock = Mock()

    @kopf.on.validate(*resource)
    def fn(**kwargs):
        mock(**kwargs)

    adm_request['request']['dryRun'] = dryrun
    await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert mock.call_count == 1
    assert mock.call_args.kwargs['dryrun'] == dryrun


async def test_headers_passed(
        settings, registry, resource, memories, insights, indices, adm_request):
    mock = Mock()

    @kopf.on.validate(*resource)
    def fn(**kwargs):
        mock(**kwargs)

    headers = {'X': '123', 'Y': '456'}
    await serve_admission_request(
        adm_request, headers=headers,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert mock.call_count == 1
    assert mock.call_args.kwargs['headers'] == headers


async def test_headers_not_passed_but_injected(
        settings, registry, resource, memories, insights, indices, adm_request):
    mock = Mock()

    @kopf.on.validate(*resource)
    def fn(**kwargs):
        mock(**kwargs)

    await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert mock.call_count == 1
    assert mock.call_args.kwargs['headers'] == {}


async def test_sslpeer_passed(
        settings, registry, resource, memories, insights, indices, adm_request):
    mock = Mock()

    @kopf.on.validate(*resource)
    def fn(**kwargs):
        mock(**kwargs)

    sslpeer = {'X': '123', 'Y': '456'}
    await serve_admission_request(
        adm_request, sslpeer=sslpeer,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert mock.call_count == 1
    assert mock.call_args.kwargs['sslpeer'] == sslpeer


async def test_sslpeer_not_passed_but_injected(
        settings, registry, resource, memories, insights, indices, adm_request):
    mock = Mock()

    @kopf.on.validate(*resource)
    def fn(**kwargs):
        mock(**kwargs)

    await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert mock.call_count == 1
    assert mock.call_args.kwargs['sslpeer'] == {}


async def test_userinfo_passed(
        settings, registry, resource, memories, insights, indices, adm_request):
    mock = Mock()

    @kopf.on.validate(*resource)
    def fn(**kwargs):
        mock(**kwargs)

    userinfo = {'X': '123', 'Y': '456'}
    adm_request['request']['userInfo'] = userinfo
    await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert mock.call_count == 1
    assert mock.call_args.kwargs['userinfo'] == userinfo


async def test_subresource_passed(
        settings, registry, resource, memories, insights, indices, adm_request):
    mock = Mock()

    @kopf.on.validate(*resource, subresource='*')
    def fn(**kwargs):
        mock(**kwargs)

    adm_request['request']['subResource'] = 'xyz'
    await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert mock.call_count == 1
    assert mock.call_args.kwargs['subresource'] == 'xyz'



"""
We test these aspects:

- If we skip the irrelevant patches: empty or not applicable (for status).
- If the status patches are executed separately if it is a subresource.
- If the patches are formed properly (ops-wise, content-changing).
- If the patches contain the "test" op with the proper `resourceVersion`.
  - Of the latest preceding patch, if any.
  - Of the original body if no patches precede it.
- If we react properly to the HTTP 422 failures of the "test" op.

We have 0-4 patches:

- merge-patch on the object
- merge-patch on the status
- json-patch on the object
- json-patch on the status

Each can be present or absent, depending on the setup:

- is status a subresource: yes or no?
- are fns present and change anything: yes or no?

Instead of doing the full matrix, we focus only on the latest patch
and the patch chronologically preceding it (for its resourceVersion),
skipping the cases when 2+ patches precede it (affects nothing),
and the cases when other patches follow it (tested in other functions).

Those are "happy paths".

Additionally, the JSON-patches can catch HTTP 422 and return
different results (the remaining patches).

Additionally, every of these patches can catch HTTP API errors
(of which HTTP 404 is handled gracefully, while others escalate).
"""
import dataclasses

import pytest

from kopf._cogs.clients.errors import APIError
from kopf._cogs.clients.patching import patch_obj
from kopf._cogs.structs.patches import Patch
from kopf._cogs.structs.references import Resource

# The key difference is the `resourceVersion` — we check that the proper one is used.
# The empty status is present to ensure predictable json-patch diffs with one field only.
OBJECT_MERGE_RESPONSE = {'metadata': {'resourceVersion': 'rv-object-merge'}, 'status': {}}
STATUS_MERGE_RESPONSE = {'metadata': {'resourceVersion': 'rv-status-merge'}, 'status': {}}
OBJECT_JSONP_RESPONSE = {'metadata': {'resourceVersion': 'rv-object-jsonp'}, 'status': {}}
STATUS_JSONP_RESPONSE = {'metadata': {'resourceVersion': 'rv-status-jsonp'}, 'status': {}}


@pytest.fixture()
def resource():
    # We do not care about namespaced/cluster-wide here, only on the subresource presence.
    return Resource('kopf.dev', 'v1', 'kopfexamples', subresources=frozenset({'status'}))


@pytest.fixture(autouse=True)
def _endpoints(kmock, resource, namespace):

    def set_body(body):
        kmock.objects[resource, namespace, 'name1'] = dict(body)
        return body

    set_body({})  # suppress 404s in the emulator

    # Every time the request is made, also update the stored body to the request's response,
    # so that json-patches could be applied properly on the next requests, in particular:
    # - the validity of he "test" operations on "/metadata/resourceVersion";
    # - the existence of the metadata/spec/status stanzas for proper "add/replace" ops.
    common = kmock['patch', resource, kmock.namespace(namespace), kmock.name('name1')]
    common[{'Content-Type': 'application/merge-patch+json'}, kmock.subresource(None)] << (lambda: set_body(OBJECT_MERGE_RESPONSE))
    common[{'Content-Type': 'application/merge-patch+json'}, kmock.subresource('status')] << (lambda: set_body(STATUS_MERGE_RESPONSE))
    common[{'Content-Type': 'application/json-patch+json'}, kmock.subresource(None)] << (lambda: set_body(OBJECT_JSONP_RESPONSE))
    common[{'Content-Type': 'application/json-patch+json'}, kmock.subresource('status')] << (lambda: set_body(STATUS_JSONP_RESPONSE))


def _noop(body):
    pass


def _add_finalizer(body):
    body.setdefault('metadata', {}).setdefault('finalizers', []).append('fin')


def _add_status_field(body):
    body.setdefault('status', {})['count'] = 1


#
# Test sequencing of patches and proper use of `resourceVersion` on each step.
# Limit to max 2 patches per operation — there is no extra logic in 3+ patches.
#
async def test_empty_or_noop_patch_makes_no_api_calls(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    patch = Patch({}, body=original_body, fns=[_noop])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )

    assert result is None
    assert remaining is None
    assert len(kmock) == 0

    assert_logs([
    ], prohibited=[
        "Merge-patching the resource with",
        "Merge-patching the status with",
        "JSON-patching the resource with",
        "JSON-patching the status with",
    ])


async def test_object_merge_alone(caplog,
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'spec': {'x': 'y'}}, body=original_body, fns=[])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == OBJECT_MERGE_RESPONSE
    assert remaining is None
    assert len(kmock) == 1
    assert kmock[0].subresource is None
    assert kmock[0].data == {'spec': {'x': 'y'}}
    assert_logs([
        "Merge-patching the resource with",
    ], prohibited=[
        "Merge-patching the status with",
        "JSON-patching the resource with",
        "JSON-patching the status with",
    ])


async def test_status_merge_alone(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'status': {'x': 'y'}}, body=original_body, fns=[])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_MERGE_RESPONSE
    assert remaining is None
    assert len(kmock) == 1
    assert kmock[0].subresource == 'status'
    assert kmock[0].data == {'status': {'x': 'y'}}
    assert_logs([
        "Merge-patching the status with",
    ], prohibited=[
        "Merge-patching the resource with",
        "JSON-patching the resource with",
        "JSON-patching the status with",
    ])


async def test_status_merge_after_object_merge(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'spec': {'x': 'y'}, 'status': {'x': 'y'}}, body=original_body, fns=[])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_MERGE_RESPONSE
    assert remaining is None
    assert len(kmock) == 2
    assert kmock[0].subresource is None
    assert kmock[1].subresource == 'status'
    assert kmock[0].data == {'spec': {'x': 'y'}}  # status absent
    assert kmock[1].data == {'status': {'x': 'y'}}  # spec absent
    assert_logs([
        "Merge-patching the resource with",
        "Merge-patching the status with",
    ], prohibited=[
        "JSON-patching the resource with",
        "JSON-patching the status with",
    ])


async def test_object_jsonp_alone(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({}, body=original_body, fns=[_add_finalizer])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == OBJECT_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 1
    assert kmock[0].subresource is None
    assert kmock[0].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv0'},
        {'op': 'add', 'path': '/metadata/finalizers', 'value': ['fin']},
    ]
    assert_logs([
        "JSON-patching the resource with",
    ], prohibited=[
        "Merge-patching the resource with",
        "Merge-patching the status with",
        "JSON-patching the status with",
    ])


async def test_object_jsonp_after_object_merge(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'spec': {'x': 'y'}}, body=original_body, fns=[_add_finalizer])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == OBJECT_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 2
    assert kmock[0].subresource is None
    assert kmock[1].subresource is None
    assert kmock[0].data == {'spec': {'x': 'y'}}
    assert kmock[1].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv-object-merge'},
        {'op': 'add', 'path': '/metadata/finalizers', 'value': ['fin']},
    ]
    assert_logs([
        "Merge-patching the resource with",
        "JSON-patching the resource with",
    ], prohibited=[
        "Merge-patching the status with",
        "JSON-patching the status with",
    ])


async def test_object_jsonp_after_status_merge(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'status': {'x': 'y'}}, body=original_body, fns=[_add_finalizer])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == OBJECT_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 2
    assert kmock[0].subresource == 'status'
    assert kmock[1].subresource is None
    assert kmock[0].data == {'status': {'x': 'y'}}
    assert kmock[1].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv-status-merge'},
        {'op': 'add', 'path': '/metadata/finalizers', 'value': ['fin']},
    ]
    assert_logs([
        "Merge-patching the status with",
        "JSON-patching the resource with",
    ], prohibited=[
        "Merge-patching the resource with",
        "JSON-patching the status with",
    ])


async def test_status_jsonp_alone(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({}, body=original_body, fns=[_add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 1
    assert kmock[0].subresource == 'status'
    assert kmock[0].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv0'},
        {'op': 'add', 'path': '/status/count', 'value': 1},
    ]
    assert_logs([
        "JSON-patching the status with",
    ], prohibited=[
        "Merge-patching the resource with",
        "Merge-patching the status with",
        "JSON-patching the resource with",
    ])


async def test_status_jsonp_after_object_merge(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'spec': {'x': 'y'}}, body=original_body, fns=[_add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 2
    assert kmock[0].subresource is None
    assert kmock[1].subresource == 'status'
    assert kmock[0].data == {'spec': {'x': 'y'}}
    assert kmock[1].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv-object-merge'},
        {'op': 'add', 'path': '/status/count', 'value': 1},
    ]
    assert_logs([
        "Merge-patching the resource with",
        "JSON-patching the status with",
    ], prohibited=[
        "Merge-patching the status with",
        "JSON-patching the resource with",
    ])


async def test_status_jsonp_after_status_merge(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'status': {'x': 'y'}}, body=original_body, fns=[_add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 2
    assert kmock[0].subresource == 'status'
    assert kmock[1].subresource == 'status'
    assert kmock[0].data == {'status': {'x': 'y'}}
    assert kmock[1].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv-status-merge'},
        {'op': 'add', 'path': '/status/count', 'value': 1},
    ]
    assert_logs([
        "Merge-patching the status with",
        "JSON-patching the status with",
    ], prohibited=[
        "Merge-patching the resource with",
        "JSON-patching the resource with",
    ])


async def test_status_jsonp_after_object_jsonp(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch(body=original_body, fns=[_add_finalizer, _add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 2
    assert kmock[0].subresource is None
    assert kmock[1].subresource == 'status'
    assert kmock[0].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv0'},
        {'op': 'add', 'path': '/metadata/finalizers', 'value': ['fin']},
    ]
    assert kmock[1].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv-object-jsonp'},
        {'op': 'add', 'path': '/status/count', 'value': 1},
    ]
    assert_logs([
        "JSON-patching the resource with",
        "JSON-patching the status with",
    ], prohibited=[
        "Merge-patching the resource with",
        "Merge-patching the status with",
    ])


# Not of direct interest (already tested implicitly), but worth checking the worst case: 4x patches.
async def test_all_four_patches(
        kmock, settings, resource, namespace, logger, assert_logs):
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'spec': {'x': 'y'}, 'status': {'x': 'y'}}, body=original_body, fns=[_add_finalizer, _add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 4
    assert kmock[0].subresource is None  # merge-patch
    assert kmock[1].subresource == 'status'  # merge-patch
    assert kmock[2].subresource is None  # json-patch
    assert kmock[3].subresource == 'status'  # json-patch
    assert kmock[0].data == {'spec': {'x': 'y'}}
    assert kmock[1].data == {'status': {'x': 'y'}}
    assert kmock[2].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv-status-merge'},
        {'op': 'add', 'path': '/metadata/finalizers', 'value': ['fin']},
    ]
    assert kmock[3].data == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv-object-jsonp'},
        {'op': 'add', 'path': '/status/count', 'value': 1},
    ]
    assert_logs([
        "Merge-patching the resource with",
        "Merge-patching the status with",
        "JSON-patching the resource with",
        "JSON-patching the status with",
    ])


#
# Test status as a direct field (not a subresource): must not make extra API patches.
#
async def test_no_subresource_skips_status_merge(
        kmock, settings, resource, namespace, logger, assert_logs):
    resource = dataclasses.replace(resource, subresources=frozenset())
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch({'spec': {'x': 'y'}, 'status': {'x': 'y'}}, body=original_body)
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == OBJECT_MERGE_RESPONSE
    assert remaining is None
    assert len(kmock) == 1
    assert kmock[0].subresource is None
    assert kmock[0].data == {'spec': {'x': 'y'}, 'status': {'x': 'y'}}  # mixed!
    assert_logs([
        "Merge-patching the resource with",
    ], prohibited=[
        "Merge-patching the status with",
        "JSON-patching the resource with",
        "JSON-patching the status with",
    ])


async def test_no_subresource_skips_status_jsonp(
        kmock, settings, resource, namespace, logger, assert_logs):
    resource = dataclasses.replace(resource, subresources=frozenset())
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    kmock.objects[resource, namespace, 'name1'] = original_body
    patch = Patch(body=original_body, fns=[_add_finalizer, _add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == OBJECT_JSONP_RESPONSE
    assert remaining is None
    assert len(kmock) == 1
    assert kmock[0].subresource is None
    assert kmock[0].data[:1] == [
        {'op': 'test', 'path': '/metadata/resourceVersion', 'value': 'rv0'},
    ]
    # The order of ops does not matter. What matters is the mix of status and metadata.
    paths = {op['path'] for op in kmock[0].data[1:]}
    assert paths == {'/metadata/finalizers', '/status/count'}  # mixed!
    assert_logs([
        "JSON-patching the resource with",
    ], prohibited=[
        "Merge-patching the resource with",
        "Merge-patching the status with",
        "JSON-patching the status with",
    ])


#
# Test the remaining patches on HTTP 422 from JSON-patches.
#
async def test_422_in_object_jsonp_returns_the_remaining_patch(
        kmock, settings, resource, namespace, logger, assert_logs):
    kmock[kmock.subresource(None), {'Content-Type': 'application/json-patch+json'}] ** 1 << 422
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    patch = Patch({'spec': {'x': 'y'}, 'status': {'x': 'y'}}, body=original_body, fns=[_add_finalizer, _add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == STATUS_MERGE_RESPONSE  # we only need to see that it is not None
    assert remaining is not None
    assert remaining._original is None  # do not carry the body through cycles
    assert list(remaining.fns) == [_add_finalizer, _add_status_field]
    assert dict(remaining) == {}
    assert len(kmock) == 3
    assert kmock[0].subresource is None  # merge-patch
    assert kmock[1].subresource == 'status'  # merge-patch
    assert kmock[2].subresource is None  # json-patch
    assert_logs([
        "Merge-patching the resource with",
        "Merge-patching the status with",
        "JSON-patching the resource with",
        "Could not apply the patch in full",
    ], prohibited=[
        "JSON-patching the status with",
    ])


async def test_422_in_status_jsonp_returns_the_remaining_patch(
        kmock, settings, resource, namespace, logger, assert_logs):
    kmock[kmock.subresource('status'), {'Content-Type': 'application/json-patch+json'}] ** 1 << 422
    original_body = {'metadata': {'resourceVersion': 'rv0'}, 'status': {}}
    patch = Patch({'spec': {'x': 'y'}, 'status': {'x': 'y'}}, body=original_body, fns=[_add_finalizer, _add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result == OBJECT_JSONP_RESPONSE  # we only need to see that it is not None
    assert remaining is not None
    assert remaining._original is None  # do not carry the body through cycles
    assert list(remaining.fns) == [_add_finalizer, _add_status_field]
    assert dict(remaining) == {}
    assert len(kmock) == 4
    assert kmock[0].subresource is None  # merge-patch
    assert kmock[1].subresource == 'status'  # merge-patch
    assert kmock[2].subresource is None  # json-patch
    assert kmock[3].subresource == 'status'  # json-patch
    assert_logs([
        "Merge-patching the resource with",
        "Merge-patching the status with",
        "JSON-patching the resource with",
        "JSON-patching the status with",
        "Could not apply the patch in full",
    ])


#
# Test API errors in every individual patch.
#
@pytest.mark.parametrize('exp_api_count, content_type, subresource', [
    pytest.param(1, 'application/merge-patch+json', None),
    pytest.param(2, 'application/merge-patch+json', 'status'),
    pytest.param(3, 'application/json-patch+json', None),
    pytest.param(4, 'application/json-patch+json', 'status'),
])
async def test_404_ignores_absent_objects(
        kmock, settings, resource, namespace, logger, content_type, subresource,
        cluster_resource, namespaced_resource, assert_logs, exp_api_count):
    kmock[kmock.subresource(subresource), {'Content-Type': content_type}] ** 1 << 404
    patch = Patch({'spec': {'x': 'y'}, 'status': {'x': 'y'}}, body={}, fns=[_add_finalizer, _add_status_field])
    result, remaining = await patch_obj(
        logger=logger, settings=settings, resource=resource,
        namespace=namespace, name='name1', patch=patch,
    )
    assert result is None
    assert remaining is None
    assert len(kmock) == exp_api_count
    assert_logs([
        "Patching was skipped: the object does not exist anymore",
    ])


# Note: 401 is wrapped into a LoginError and is tested elsewhere.
@pytest.mark.parametrize('exp_api_count, content_type, subresource', [
    pytest.param(1, 'application/merge-patch+json', None),
    pytest.param(2, 'application/merge-patch+json', 'status'),
    pytest.param(3, 'application/json-patch+json', None),
    pytest.param(4, 'application/json-patch+json', 'status'),
])
@pytest.mark.parametrize('status', [400, 403, 500, 666])
async def test_api_errors_raised(
        kmock, settings, status, resource, namespace, logger, content_type, subresource,
        cluster_resource, namespaced_resource, exp_api_count):
    kmock[kmock.subresource(subresource), {'Content-Type': content_type}] ** 1 << status
    patch = Patch({'spec': {'x': 'y'}, 'status': {'x': 'y'}}, body={}, fns=[_add_finalizer, _add_status_field])
    with pytest.raises(APIError) as e:
        await patch_obj(
            logger=logger, settings=settings, resource=resource,
            namespace=namespace, name='name1', patch=patch,
        )
    assert len(kmock) == exp_api_count
    assert e.value.status == status



import asyncio
import base64
import json
import ssl

import aiohttp
import pytest

from kopf._core.engines.admission import AmbiguousResourceError, MissingDataError, \
                                         UnknownResourceError, WebhookError
from kopf._kits.webhooks import WebhookDockerDesktopServer, WebhookK3dServer, \
                                WebhookMinikubeServer, WebhookServer


async def test_starts_as_http_ipv4(responder):
    server = WebhookServer(addr='127.0.0.1', port=22533, path='/p1/p2', insecure=True)
    async with server:
        async for client_config in server(responder.fn):
            assert client_config['url'] == 'http://127.0.0.1:22533/p1/p2'
            assert 'caBundle' not in client_config
            break  # do not sleep


async def test_starts_as_http_ipv6(responder):
    server = WebhookServer(addr='::1', port=22533, path='/p1/p2', insecure=True)
    async with server:
        async for client_config in server(responder.fn):
            assert client_config['url'] == 'http://[::1]:22533/p1/p2'
            assert 'caBundle' not in client_config
            break  # do not sleep


async def test_unspecified_port_allocates_a_random_port(responder):
    server1 = WebhookServer(addr='127.0.0.1', path='/p1/p2', insecure=True)
    server2 = WebhookServer(addr='127.0.0.1', path='/p1/p2', insecure=True)
    async with server1, server2:
        async for client_config1 in server1(responder.fn):
            async for client_config2 in server2(responder.fn):
                assert client_config1['url'] != client_config2['url']
                break  # do not sleep
            break  # do not sleep


async def test_unspecified_addr_uses_all_interfaces(responder, assert_logs):
    server = WebhookServer(port=22533, path='/p1/p2', insecure=True)
    async with server:
        async for client_config in server(responder.fn):
            assert client_config['url'] == 'http://localhost:22533/p1/p2'
            break  # do not sleep
    assert_logs([r"Listening for webhooks at http://\*:22533/p1/p2"])


async def test_webhookserver_starts_as_https_with_selfsigned_cert(
        responder):
    server = WebhookServer(addr='127.0.0.1', port=22533, path='/p1/p2', host='somehost')
    async with server:
        async for client_config in server(responder.fn):
            assert client_config['url'] == 'https://somehost:22533/p1/p2'
            assert 'caBundle' in client_config  # regardless of the value
            break  # do not sleep


async def test_webhookserver_starts_as_https_with_provided_cert(
        certfile, pkeyfile, certpkey, responder):
    server = WebhookServer(port=22533, certfile=certfile, pkeyfile=pkeyfile)
    async with server:
        async for client_config in server(responder.fn):
            assert client_config['url'] == 'https://localhost:22533'
            assert base64.b64decode(client_config['caBundle']) == certpkey[0]
            break  # do not sleep


@pytest.mark.parametrize('cls, url', [
    (WebhookK3dServer, 'https://host.k3d.internal:22533/p1/p2'),
    (WebhookMinikubeServer, 'https://host.minikube.internal:22533/p1/p2'),
    (WebhookDockerDesktopServer, 'https://host.docker.internal:22533/p1/p2'),
])
async def test_webhookserver_flavours_inject_hostnames(
        certfile, pkeyfile, certpkey, responder, cls, url):
    server = cls(port=22533, certfile=certfile, pkeyfile=pkeyfile, path='/p1/p2')
    async with server:
        async for client_config in server(responder.fn):
            assert client_config['url'] == url
            break  # do not sleep


@pytest.mark.usefixtures('no_sslproto_warnings')
async def test_webhookserver_serves(
        certfile, pkeyfile, responder, adm_request):
    responder.fut.set_result({'hello': 'world'})
    server = WebhookServer(certfile=certfile, pkeyfile=pkeyfile)
    async with server:
        async for client_config in server(responder.fn):
            cadata = base64.b64decode(client_config['caBundle']).decode('ascii')
            sslctx = ssl.create_default_context(cadata=cadata)
            async with aiohttp.ClientSession() as client:
                async with client.post(client_config['url'], ssl=sslctx, json=adm_request) as resp:
                    text = await resp.text()
                    assert text == '{"hello": "world"}'
                    assert resp.status == 200
            break  # do not sleep


async def test_webhookserver_renews_certificate_files(tmp_path, responder, looptime):

    # Generate different certificates.
    certpath = tmp_path / 'cert.pem'
    pkeypath = tmp_path / 'pkey.pem'
    cert1, pkey1 = WebhookServer.build_certificate(['localhost', '127.0.0.1', 'hostA'])
    cert2, pkey2 = WebhookServer.build_certificate(['localhost', '127.0.0.1', 'hostB'])
    certpath.write_bytes(cert1)
    pkeypath.write_bytes(pkey1)

    # Schedule a delayed replacement of the files.
    loop = asyncio.get_running_loop()
    loop.call_later(5, certpath.write_bytes, cert2)
    loop.call_later(5, pkeypath.write_bytes, pkey2)

    # Run the server.
    cabundles = []
    server = WebhookServer(certfile=certpath, pkeyfile=pkeypath, file_check_interval=10)
    async with server:
        async for client_config in server(responder.fn):
            cabundles.append(base64.b64decode(client_config['caBundle']))
            if len(cabundles) >= 2:
                break

    assert looptime == 10
    assert len(cabundles) == 2
    assert cabundles[0] == cert1
    assert cabundles[1] == cert2


@pytest.mark.parametrize('code, error', [
    (500, Exception),
    (400, WebhookError),
    (400, WebhookError),
    (400, MissingDataError),
    (404, UnknownResourceError),
    (409, AmbiguousResourceError),
    (400, lambda: json.JSONDecodeError('...', '...', 0)),
])
@pytest.mark.usefixtures('no_sslproto_warnings')
async def test_webhookserver_errors(
        certfile, pkeyfile, responder, adm_request, code, error):
    responder.fut.set_exception(error())
    server = WebhookServer(certfile=certfile, pkeyfile=pkeyfile)
    async with server:
        async for client_config in server(responder.fn):
            cadata = base64.b64decode(client_config['caBundle']).decode('ascii')
            sslctx = ssl.create_default_context(cadata=cadata)
            async with aiohttp.ClientSession() as client:
                async with client.post(client_config['url'], ssl=sslctx, json=adm_request) as resp:
                    assert resp.status == code
            break  # do not sleep



"""
Admission reviews: requests & responses, also the webhook server protocols.
"""
from collections.abc import AsyncIterator, Awaitable, Callable
from typing import Any, Literal, Protocol, TypedDict

from kopf._cogs.structs import bodies

Headers = dict[str, str]
SSLPeer = dict[str, Any]

Operation = Literal['CREATE', 'UPDATE', 'DELETE', 'CONNECT']


class RequestKind(TypedDict):
    group: str
    version: str
    kind: str


class RequestResource(TypedDict):
    group: str
    version: str
    resource: str


class UserInfo(TypedDict):
    username: str
    uid: str
    groups: list[str]


class CreateOptions(TypedDict, total=False):
    apiVersion: Literal["meta.k8s.io/v1"]
    kind: Literal["CreateOptions"]


class UpdateOptions(TypedDict, total=False):
    apiVersion: Literal["meta.k8s.io/v1"]
    kind: Literal["UpdateOptions"]


class DeleteOptions(TypedDict, total=False):
    apiVersion: Literal["meta.k8s.io/v1"]
    kind: Literal["DeleteOptions"]


class RequestPayload(TypedDict):
    uid: str
    kind: RequestKind
    resource: RequestResource
    subResource: str | None
    requestKind: RequestKind
    requestResource: RequestResource
    requestSubResource: str | None
    userInfo: UserInfo
    name: str
    namespace: str | None
    operation: Operation
    options: CreateOptions | UpdateOptions | DeleteOptions | None
    dryRun: bool
    object: bodies.RawBody
    oldObject: bodies.RawBody | None


class Request(TypedDict):
    apiVersion: Literal["admission.k8s.io/v1", "admission.k8s.io/v1beta1"]
    kind: Literal["AdmissionReview"]
    request: RequestPayload


class ResponseStatus(TypedDict, total=False):
    code: int
    message: str


class ResponsePayload(TypedDict, total=False):
    uid: str
    allowed: bool
    warnings: list[str] | None
    status: ResponseStatus | None
    patch: str | None
    patchType: Literal["JSONPatch"] | None


class Response(TypedDict):
    apiVersion: Literal["admission.k8s.io/v1", "admission.k8s.io/v1beta1"]
    kind: Literal["AdmissionReview"]
    response: ResponsePayload


class WebhookClientConfigService(TypedDict, total=False):
    namespace: str | None
    name: str | None
    path: str | None
    port: int | None


class WebhookClientConfig(TypedDict, total=False):
    """
    A config of clients (apiservers) to access the webhooks' server (operators).

    This dictionary is put into managed webhook configurations "as is".
    The fields & type annotations are only for hinting.

    Kopf additionally modifies the url and the service's path to inject
    handler ids as the last path component. This must be taken into account
    by custom webhook servers.
    """
    caBundle: str | None  # if absent, the default apiservers' trust chain is used.
    url: str | None
    service: WebhookClientConfigService | None


class WebhookFn(Protocol):
    """
    A framework-provided function to call when an admission request is received.

    The framework provides the actual function. Custom webhook servers must
    accept the function, invoke it accordingly on admission requests, wait
    for the admission response, serialise it and send it back. They do not
    implement this function. This protocol only declares the exact signature.
    """
    def __call__(
            self,
            request: Request,
            *,
            webhook: str | None = None,
            headers: Headers | None = None,
            sslpeer: SSLPeer | None = None,
    ) -> Awaitable[Response]:
        ...


# A server (either a coroutine or a callable object).
WebhookServerProtocol = Callable[[WebhookFn], AsyncIterator[WebhookClientConfig]]



import dataclasses
from unittest.mock import Mock

import pytest

from kopf._cogs.structs.bodies import Body
from kopf._cogs.structs.dicts import parse_field
from kopf._cogs.structs.references import Resource
from kopf._core.intents.causes import WatchingCause
from kopf._core.intents.handlers import WatchingHandler
from kopf._core.intents.registries import match, prematch


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn(x=None):
    pass


@pytest.fixture()
def callback():
    mock = Mock()
    mock.return_value = True
    return mock


@pytest.fixture(params=['annotations', 'labels', 'value', 'when'])
def handler(request, callback, selector):
    handler = WatchingHandler(
        selector=selector,
        annotations={'known': 'value'},
        labels={'known': 'value'},
        field=parse_field('spec.field'),
        value='value',
        when=None,
        fn=some_fn, id='a', param=None, errors=None, timeout=None, retries=None, backoff=None,
    )
    if request.param in ['annotations', 'labels']:
        handler = dataclasses.replace(handler, **{request.param: {'known': callback}})
    else:
        handler = dataclasses.replace(handler, **{request.param: callback})
    return handler


@pytest.fixture()
def cause(cause_factory, callback):
    return cause_factory(
        cls=WatchingCause,
        body=Body(dict(
            metadata=dict(
                labels={'known': 'value'},
                annotations={'known': 'value'},
            ),
            spec=dict(
                field='value',
            ),
        )))


@pytest.mark.parametrize('match_fn', [match, prematch])
def test_callback_is_called_with_matching_resource(
        match_fn, callback, handler, cause,
):
    result = match_fn(handler=handler, cause=cause)
    assert result
    assert callback.called


@pytest.mark.parametrize('match_fn', [match, prematch])
def test_callback_is_not_called_with_mismatching_resource(
        match_fn, callback, handler, cause,
):
    cause = dataclasses.replace(cause, resource=Resource(group='x', version='y', plural='z'))
    result = match_fn(handler=handler, cause=cause)
    assert not result
    assert not callback.called



import pytest

from kopf._cogs.clients.errors import APIError, APITooManyRequestsError
from kopf._cogs.clients.watching import Bookmark, infinite_watch

STREAM_WITH_UNKNOWN_EVENT = (
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'UNKNOWN', 'object': {}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
)
STREAM_WITH_ERROR_410GONE = (
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 410}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
)


class SampleException(Exception):
    pass


async def test_exception_escalates(kmock, settings, resource, namespace, enforced_session, mocker):
    enforced_session.request = mocker.Mock(side_effect=SampleException())
    kmock['watch', resource, kmock.namespace(namespace)] << ()

    events = []
    with pytest.raises(SampleException):
        async for event in infinite_watch(settings=settings,
                                          resource=resource,
                                          namespace=namespace,
                                          _iterations=1):
            events.append(event)

    assert len(events) == 0


async def test_infinite_watch_never_exits_normally(kmock, settings, resource, namespace):
    kmock['watch', resource, kmock.namespace(namespace)] << iter(STREAM_WITH_ERROR_410GONE)
    kmock['watch', resource, kmock.namespace(namespace)] << iter(STREAM_WITH_UNKNOWN_EVENT)
    kmock['watch', resource, kmock.namespace(namespace)] << 555

    events = []
    with pytest.raises(APIError) as e:
        async for event in infinite_watch(settings=settings,
                                          resource=resource,
                                          namespace=namespace):
            events.append(event)

    assert e.value.status == 555

    assert len(events) == 5
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert events[2] == Bookmark.LISTED
    assert events[3]['object']['spec'] == 'a'
    assert events[4]['object']['spec'] == 'b'


async def test_too_many_requests_exception_old_style(
        kmock, settings, resource, namespace, enforced_session, mocker):

    exc = APITooManyRequestsError({
        "apiVersion": "v1",
        "code": 666,  # must be irrelevant, purely informational
        "status": "Failure",
        "details": {
            "retryAfterSeconds": 1,
        }
    }, status=429, headers={})
    enforced_session.request = mocker.Mock(side_effect=exc)
    kmock['watch', resource, kmock.namespace(namespace)] << ()

    events = []
    async for event in infinite_watch(settings=settings,
                                      resource=resource,
                                      namespace=namespace,
                                      _iterations=1):
        events.append(event)

    assert len(events) == 0


async def test_too_many_requests_exception_new_style(
        settings, resource, kmock, namespace, enforced_session, mocker):

    headers = {'Retry-After': '1'}
    exc = APITooManyRequestsError("Too many requests.", status=429, headers=headers)
    enforced_session.request = mocker.Mock(side_effect=exc)
    kmock['watch', resource, kmock.namespace(namespace)] << ()

    events = []
    async for event in infinite_watch(settings=settings,
                                      resource=resource,
                                      namespace=namespace,
                                      _iterations=1):
        events.append(event)

    assert len(events) == 0



"""
All the structures needed for Kubernetes patching.

Currently, it is implemented via a JSON merge-patch (RFC 7386),
i.e. a simple dictionary with field overrides, and ``None`` for field deletions.

In the future, it can be extended to a standalone object, which exposes
a dict-like behavior, and remembers the changes in order of their execution,
and then generates the JSON patch (RFC 6902).
"""
import collections.abc
import copy
from collections.abc import Callable, Iterable
from typing import Any, Literal, TypedDict, cast

import jsonpatch

from kopf._cogs.structs import bodies, dicts

JSONPatchOp = Literal["add", "replace", "remove", "test", "move", "copy"]


class JSONPatchItem(TypedDict, total=False):
    op: JSONPatchOp
    # from: str  # but it is a python keyword
    path: str
    value: Any | None


JSONPatch = list[JSONPatchItem]

# An arbitrary transformation function, which modifies the body in place.
PatchFn = Callable[[bodies.RawBody], None]


class MetaPatch(dicts.MutableMappingView[str, Any]):
    _labels: dicts.MutableMappingView[str, str | None]
    _annotations: dicts.MutableMappingView[str, str | None]

    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'metadata')
        self._labels = dicts.MutableMappingView(self, 'labels')
        self._annotations = dicts.MutableMappingView(self, 'annotations')

    @property
    def labels(self) -> dicts.MutableMappingView[str, str | None]:
        return self._labels

    @property
    def annotations(self) -> dicts.MutableMappingView[str, str | None]:
        return self._annotations


class SpecPatch(dicts.MutableMappingView[str, Any]):
    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'spec')


class StatusPatch(dicts.MutableMappingView[str, Any]):
    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'status')


# Event-handling structures, used internally in the framework and handlers only.
class Patch(dict[str, Any]):

    def __init__(
        self,
        src: dict[str, Any] | None = None,
        /,
        body: bodies.RawBody | None = None,
        fns: Iterable[PatchFn] = (),
    ) -> None:
        super().__init__(src or {})
        self._meta = MetaPatch(self)
        self._spec = SpecPatch(self)
        self._status = StatusPatch(self)
        self._original = body
        self._fns = (src.fns if isinstance(src, Patch) else []) + list(fns)

    def __repr__(self) -> str:
        texts: list[str] = []
        if list(self):  # any keys at all?
            texts += [super().__repr__()]
        if self.fns:
            texts += [f"fns={list(self.fns)!r}"]
        text = ", ".join(texts)
        return f"{type(self).__name__}({text})"

    def __bool__(self) -> bool:
        return len(self) > 0 or bool(self.fns)

    @property
    def fns(self) -> list[PatchFn]:
        return self._fns

    @property
    def metadata(self) -> MetaPatch:
        return self._meta

    @property
    def meta(self) -> MetaPatch:
        return self._meta

    @property
    def spec(self) -> SpecPatch:
        return self._spec

    @property
    def status(self) -> StatusPatch:
        return self._status

    def as_json_patch(self, body: bodies.RawBody | None = None) -> JSONPatch:
        """
        Build a list of JSON-patch ops for the changes & transformations.

        As a reference resource body, either the argument is used (if provided),
        or the original resource body. But the reference body is mandatory ---
        the patch calculates the differences relative to the reference body.

        Some changes might disappear from the list if they are useless (no-op):
        e.g., setting a key to ``None`` to delete it when it is already absent;
        or setting the key to a value which is already in the resource body.
        """
        # Clone the original body to be mutated in memory before diffing.
        body_as_is = body if body is not None else self._original
        body_to_be = copy.deepcopy(body_as_is)
        if not self:
            return []
        if body_as_is is None or body_to_be is None:
            raise ValueError("Cannot build a JSON-patch without the original body as a reference.")

        # Apply the changes: merge-patches first (since they are not smart to be the last ones).
        # Then all callable transformations on top of the mutated body.
        self._apply_patch(body_to_be, (), dict(self))
        for fn in self.fns:
            fn(body_to_be)

        # Calculate the actual JSON ops for this particular state of the resource.
        # No "test" operations in pure JSON-patches as used in the mutating admission calls.
        ops: JSONPatch = jsonpatch.JsonPatch.from_diff(dict(body_as_is), dict(body_to_be)).patch
        return ops

    def _apply_patch(self, body: bodies.RawBody, path: dicts.FieldPath, value: object) -> None:
        """Apply the merge-patch instructions to the mutable raw body."""
        # TODO: LATER: optimize: we now dive into the dict for every key in dicts.ensure(),
        #       but we can merge the dict + patch in one run — implement such a function.
        #       priority: low — patches are usually small, so not a big problem for now.
        match value:
            case None:
                dicts.remove(cast(dict[Any, Any], body), path)
            case collections.abc.Mapping():
                for key, val in value.items():
                    self._apply_patch(body, path + (key,), val)
            case _:
                # NB: lists overwrite the whole value, as the merge-patch does; no strategic merges.
                dicts.ensure(cast(dict[Any, Any], body), path, value)



import pytest

from kopf._cogs.structs.references import CLUSTER_PEERINGS_K, CLUSTER_PEERINGS_Z, \
                                          NAMESPACED_PEERINGS_K, NAMESPACED_PEERINGS_Z
from kopf._core.engines.peering import guess_selectors


@pytest.mark.parametrize('namespaced, expected_selectors', [
    (False, [CLUSTER_PEERINGS_K, CLUSTER_PEERINGS_Z]),
    (True, [NAMESPACED_PEERINGS_K, NAMESPACED_PEERINGS_Z]),
])
@pytest.mark.parametrize('mandatory', [False, True])
def test_resource_when_not_standalone(settings, namespaced, mandatory, expected_selectors):
    settings.peering.standalone = False
    settings.peering.namespaced = namespaced
    settings.peering.mandatory = mandatory
    selectors = guess_selectors(settings=settings)
    assert selectors == expected_selectors


@pytest.mark.parametrize('namespaced', [False, True])
@pytest.mark.parametrize('mandatory', [False, True])
def test_resource_when_standalone(settings, namespaced, mandatory):
    settings.peering.standalone = True
    settings.peering.namespaced = namespaced
    settings.peering.mandatory = mandatory
    selectors = guess_selectors(settings=settings)
    assert not selectors



import asyncio
from unittest.mock import DEFAULT

import pytest


@pytest.fixture()
def incremental_stream(kmock, resource, namespace):
    list_data = {'items': [], 'metadata': {'resourceVersion': '0'}}
    kmock['list', resource, kmock.namespace(namespace)] << list_data

    kmock['watch', resource, kmock.namespace(namespace), kmock.params(resourceVersion=0)] << (
        {'type': 'ADDED', 'object': {'metadata': {'resourceVersion': '1'}, 'spec': {'field': 'a'}}},
    )
    kmock['watch', resource, kmock.namespace(namespace), kmock.params(resourceVersion=1)] << (
        {'type': 'ADDED', 'object': {'metadata': {'resourceVersion': '2'}, 'spec': {'field': 'b'}}},
    )
    kmock['watch', resource, kmock.namespace(namespace), kmock.params(resourceVersion=2)] << (
        {'type': 'ADDED', 'object': {'metadata': {'resourceVersion': '3'}, 'spec': {'field': 'b'}}},
    )


@pytest.mark.usefixtures('incremental_stream', 'watcher_in_background')
async def test_consistency_tracking_in_the_watcher(settings, processor):

    # Override the default timeouts to make the tests faster.
    settings.queueing.idle_timeout = 100  # should not be involved, fail if it is
    settings.queueing.exit_timeout = 100  # should exit instantly, fail if it didn't
    settings.persistence.consistency_timeout = 3.45
    processed = asyncio.Event()

    def _processor(**_):
        nonlocal processed
        processed.set()
        return DEFAULT

    processor.side_effect = _processor

    # Stage 1: simulate the patching and setting the "expected" ResourceVersion.
    processor.return_value = '3'
    processed.clear()
    await processed.wait()
    assert processor.call_count == 1
    assert processor.call_args.kwargs['raw_event']['object']['metadata']['resourceVersion'] == '1'
    assert processor.call_args.kwargs['consistency_time'] is None

    # Stage 2: simulate an event with an INCONSISTENT ResourceVersion, which should be ignored.
    processor.return_value = None
    processed.clear()
    await processed.wait()
    assert processor.call_count == 2
    assert processor.call_args.kwargs['raw_event']['object']['metadata']['resourceVersion'] == '2'
    assert processor.call_args.kwargs['consistency_time'] == 3.45

    # Stage 3: simulate an event with a CONSISTENT ResourceVersion, which should be fully processed.
    processor.return_value = None
    processed.clear()
    await processed.wait()
    assert processor.call_count == 3
    assert processor.call_args.kwargs['raw_event']['object']['metadata']['resourceVersion'] == '3'
    assert processor.call_args.kwargs['consistency_time'] is None



import asyncio
import functools
from unittest.mock import AsyncMock

import pytest

from kopf._cogs.clients.watching import infinite_watch
from kopf._core.reactor.queueing import watcher, worker as original_worker

STREAM_WITH_ERROR_410GONE_ONLY = (
    {'type': 'ERROR', 'object': {'code': 410}},
)


@pytest.fixture(autouse=True)
def _enforced_api_server(fake_vault, enforced_session, resource):
    pass


@pytest.fixture()
def processor():
    """ A mock for processor---to be checked if the handler has been called. """
    return AsyncMock(return_value=None)


@pytest.fixture()
def worker_spy(mocker):
    """ Spy on the watcher: actually call it, but provide the mock-fields. """
    spy = AsyncMock(spec=original_worker, wraps=original_worker)
    return mocker.patch('kopf._core.reactor.queueing.worker', spy)


@pytest.fixture()
def worker_mock(mocker):
    """ Prevent the queue consumption, so that the queues could be checked. """
    return mocker.patch('kopf._core.reactor.queueing.worker')


@pytest.fixture()
def watcher_limited(kmock, mocker, settings):
    """ Make event streaming finite, watcher exits after depletion. """
    settings.watching.reconnect_backoff = 0
    mocker.patch('kopf._cogs.clients.watching.infinite_watch',
                 new=functools.partial(infinite_watch, _iterations=1))

    # Also ensure that any watches DO terminate the infinite/continuous watch-stream
    # when there are no explicitly added reactions left.
    (kmock['watch'] ** -50) << STREAM_WITH_ERROR_410GONE_ONLY


@pytest.fixture()
async def watcher_in_background(settings, resource, worker_spy, kmock, namespace, processor):

    # Prevent any real streaming for the very beginning, before it even starts.
    kmock['watch', resource] << ()

    # Spawn a watcher in the background.
    coro = watcher(
        namespace=namespace,
        resource=resource,
        settings=settings,
        processor=processor,
    )
    task = asyncio.create_task(coro)

    try:
        # Go for a test.
        yield
    finally:
        # Terminate the watcher to cleanup the loop.
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass  # cancellations are expected at this point



import aiohttp
import pytest

from kopf._cogs.clients.auth import APIContext, authenticated
from kopf._cogs.clients.errors import APIClientError, APIConflictError, APIError, \
                                      APIForbiddenError, APINotFoundError, \
                                      APIServerError, APITooManyRequestsError, \
                                      APIUnprocessableEntityError, check_response


@authenticated
async def get_it(url: str, *, context: APIContext) -> None:
    response = await context.session.get(url)
    await check_response(response)
    return await response.json()


def test_aiohttp_is_not_leaked_outside():
    assert not issubclass(APIError, aiohttp.ClientError)


def test_exception_without_payload():
    exc = APIError(status=456, headers={'X-H': 'abc'})
    assert exc.status == 456
    assert exc.code is None
    assert exc.message is None
    assert exc.details is None
    assert str(exc) == ""
    assert repr(exc) == "APIError(status=456)"  # no headers!


def test_exception_with_dict_payload():
    exc = APIError({"message": "msg", "code": 123, "details": {"a": "b"}}, status=456, headers={'X-H': 'abc'})
    assert exc.status == 456
    assert exc.code == 123
    assert exc.message == "msg"
    assert exc.details == {"a": "b"}
    assert str(exc) == "('msg', {'message': 'msg', 'code': 123, 'details': {'a': 'b'}})"
    assert repr(exc) == "APIError('msg', {'message': 'msg', 'code': 123, 'details': {'a': 'b'}}, status=456)"


def test_exception_with_text_payload():
    exc = APIError("oops!", status=456, headers={'X-H': 'abc'})
    assert exc.status == 456
    assert exc.code is None
    assert exc.message is None
    assert exc.details is None
    assert str(exc) == "oops!"
    assert repr(exc) == "APIError('oops!', status=456)"  # no headers!


@pytest.mark.parametrize('status', [200, 202, 300, 304])
async def test_no_error_on_success(kmock, status):
    headers = {'Content-Type': 'application/json'}
    kmock['/'] << status << headers << {"kind": "Status", "code": "xxx", "message": "msg"}
    await get_it(str(kmock.url))


# Note: 401 is wrapped into a LoginError and is tested elsewhere.
@pytest.mark.parametrize('status, exctype', [
    (403, APIForbiddenError),
    (404, APINotFoundError),
    (409, APIConflictError),
    (422, APIUnprocessableEntityError),
    (429, APITooManyRequestsError),
    (400, APIClientError),
    (403, APIClientError),
    (404, APIClientError),
    (422, APIClientError),
    (500, APIServerError),
    (503, APIServerError),
    (400, APIError),
    (500, APIError),
    (666, APIError),
])
async def test_error_with_dict_payload(kmock, status, exctype):
    headers = {'Content-Type': 'application/json'}
    kmock['/'] << status << headers << {"kind": "Status", "code": 123, "message": "msg", "details": {"a": "b"}}

    with pytest.raises(APIError) as err:
        await get_it(str(kmock.url))

    assert not isinstance(err.value, aiohttp.ClientResponseError)
    assert isinstance(err.value, exctype)
    assert err.value.status == status
    assert err.value.code == 123
    assert err.value.message == 'msg'
    assert err.value.details == {'a': 'b'}


@pytest.mark.parametrize('status', [400, 429, 500, 666])
async def test_error_with_text_payload(kmock, status):
    headers = {'Content-Type': 'application/json'}
    kmock['/'] << status << headers << b'unparsable json'

    with pytest.raises(APIError) as err:
        await get_it(str(kmock.url))

    assert err.value.status == status
    assert err.value.code is None
    assert err.value.message is None
    assert err.value.details is None
    assert str(err.value) == "unparsable json"


@pytest.mark.parametrize('status', [400, 500, 666])
async def test_error_with_parseable_nonk8s_payload(kmock, status):
    headers = {'Content-Type': 'application/json'}
    kmock['/'] << status << headers << {"kind": "NonStatus", "code": "xxx", "message": "msg"}

    with pytest.raises(APIError) as err:
        await get_it(str(kmock.url))

    assert err.value.status == status
    assert err.value.code is None
    assert err.value.message is None
    assert err.value.details is None
    assert str(err.value) == ""


async def test_cutting_the_text_response_overflow(kmock):
    kmock['get /'] << 400 << {'Content-Type': 'application/json'} << 'helloworld'*1000

    with pytest.raises(APIError) as err:
        await get_it(kmock.url)

    # 256 comes from TEXT_ERROR_MAX_SIZE, 10 is the length of "helloworld".
    assert str(err.value) == "helloworld" * (256//10) + "hel..."



import asyncio
import base64
import logging
import os.path
import pathlib
import ssl

import aiohttp
import pytest
import pytest_asyncio
from _proxy_server import ProxyServer

from kopf._cogs.clients.api import post
from kopf._cogs.clients.auth import APIContext, authenticated, vault_var
from kopf._cogs.structs.credentials import AiohttpSession, ConnectionInfo, Vault
from kopf._kits.webhooks import WebhookServer

# These are Minikube's locally generated certificates (CN=minikubeCA).
# They are not in any public use, and are regenerated regularly.
with open(os.path.join(os.path.dirname(__file__), 'fixtures/ca.pem')) as f:
    SAMPLE_MINIKUBE_CA = f.read()

with open(os.path.join(os.path.dirname(__file__), 'fixtures/cert.pem')) as f:
    SAMPLE_MINIKUBE_CERT = f.read()

with open(os.path.join(os.path.dirname(__file__), 'fixtures/pkey.pem')) as f:
    SAMPLE_MINIKUBE_PKEY = f.read()


@authenticated
async def fn(context: APIContext):
    return context.session


@pytest.fixture(autouse=True)
def vault():
    vault = Vault()
    vault_var.set(vault)
    return vault


async def test_basic_auth(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            username='username',
            password='password',
        ),
    })
    session = await fn()

    async with session:
        assert session._default_auth.login == 'username'
        assert session._default_auth.password == 'password'
        assert 'Authorization' not in session._default_headers


async def test_header_with_token_only(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            token='token',
        ),
    })
    session = await fn()

    async with session:
        assert session._default_auth is None
        assert session._default_headers['Authorization'] == 'Bearer token'


async def test_header_with_schema_only(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            scheme='Digest xyz',
        ),
    })
    session = await fn()

    async with session:
        assert session._default_auth is None
        assert session._default_headers['Authorization'] == 'Digest xyz'


async def test_header_with_schema_and_token(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            scheme='Digest',
            token='xyz',
        ),
    })
    session = await fn()

    async with session:
        assert session._default_auth is None
        assert session._default_headers['Authorization'] == 'Digest xyz'


async def test_ca_insecure(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            insecure=True,
        ),
    })
    session = await fn()

    async with session:
        ctx = session.connector._ssl
        assert ctx.verify_mode == ssl.CERT_NONE


# TODO: find a way to test that the client certificates/pkeys are indeed loaded.
# TODO: currently, we only test that the parsing/loading does not fail at all.
@pytest.mark.parametrize('pkey_path', [
    pytest.param(pathlib.Path(__file__).parent / 'fixtures' / 'pkey.pem', id='pkey-path'),
    pytest.param(os.path.join(os.path.dirname(__file__), 'fixtures/pkey.pem'), id='pkey-str'),
    pytest.param(os.path.join(os.path.dirname(__file__), 'fixtures/pkey.pem').encode(), id='pkey-bytes'),
])
@pytest.mark.parametrize('cert_path', [
    pytest.param(pathlib.Path(__file__).parent / 'fixtures' / 'cert.pem', id='cert-path'),
    pytest.param(os.path.join(os.path.dirname(__file__), 'fixtures/cert.pem'), id='cert-str'),
    pytest.param(os.path.join(os.path.dirname(__file__), 'fixtures/cert.pem').encode(), id='cert-bytes'),
])
@pytest.mark.parametrize('ca_path', [
    pytest.param(pathlib.Path(__file__).parent / 'fixtures' / 'ca.pem', id='ca-path'),
    pytest.param(os.path.join(os.path.dirname(__file__), 'fixtures/ca.pem'), id='ca-str'),
    pytest.param(os.path.join(os.path.dirname(__file__), 'fixtures/ca.pem').encode(), id='ca-bytes'),
])
async def test_ssl_context_from_path(vault, ca_path, cert_path, pkey_path):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            ca_path=ca_path,
            certificate_path=cert_path,
            private_key_path=pkey_path,
        ),
    })
    session = await fn()
    async with session:
        # We can only get the CA, not the individual certs/pkeys, so we trust it is parsed properly.
        # The functionality of the real cert/pkeys is tested in the e2e tests with the real cluster.
        ca = session.connector._ssl.get_ca_certs()
        assert ca[0]['issuer'] == ((('commonName', 'minikubeCA'),),)
        assert ca[0]['subject'] == ((('commonName', 'minikubeCA'),),)
        assert ca[0]['notAfter'] == 'May 19 09:18:36 2029 GMT'
        assert ca[0]['notBefore'] == 'May 21 09:18:36 2019 GMT'


@pytest.mark.parametrize('pkey_data', [
    pytest.param(SAMPLE_MINIKUBE_PKEY, id='pkey-raw-str'),
    pytest.param(SAMPLE_MINIKUBE_PKEY.encode(), id='pkey-raw-bytes'),
    pytest.param(base64.encodebytes(SAMPLE_MINIKUBE_PKEY.encode()), id='pkey-b64-bytes'),
    pytest.param(base64.encodebytes(SAMPLE_MINIKUBE_PKEY.encode()).decode(), id='pkey-b64-str'),
])
@pytest.mark.parametrize('cert_data', [
    pytest.param(SAMPLE_MINIKUBE_CERT, id='cert-raw-str'),
    pytest.param(SAMPLE_MINIKUBE_CERT.encode(), id='cert-raw-bytes'),
    pytest.param(base64.encodebytes(SAMPLE_MINIKUBE_CERT.encode()), id='cert-b64-bytes'),
    pytest.param(base64.encodebytes(SAMPLE_MINIKUBE_CERT.encode()).decode(), id='cert-b64-str'),
])
@pytest.mark.parametrize('ca_data', [
    pytest.param(SAMPLE_MINIKUBE_CA, id='ca-raw-str'),
    pytest.param(SAMPLE_MINIKUBE_CA.encode(), id='ca-raw-bytes'),
    pytest.param(base64.encodebytes(SAMPLE_MINIKUBE_CA.encode()), id='ca-b64-bytes'),
    pytest.param(base64.encodebytes(SAMPLE_MINIKUBE_CA.encode()).decode(), id='ca-b64-str'),
])
async def test_ssl_context_from_data(vault, ca_data, cert_data, pkey_data):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            ca_data=ca_data,
            certificate_data=cert_data,
            private_key_data=pkey_data,
        ),
    })
    session = await fn()
    async with session:
        # We can only get the CA, not the individual certs/pkeys, so we trust it is parsed properly.
        # The functionality of the real cert/pkeys is tested in the e2e tests with the real cluster.
        ca = session.connector._ssl.get_ca_certs()
        assert ca[0]['issuer'] == ((('commonName', 'minikubeCA'),),)
        assert ca[0]['subject'] == ((('commonName', 'minikubeCA'),),)
        assert ca[0]['notAfter'] == 'May 19 09:18:36 2029 GMT'
        assert ca[0]['notBefore'] == 'May 21 09:18:36 2019 GMT'


async def test_custom_aiohttp_session(vault):
    sample_session = aiohttp.ClientSession()
    await vault.populate({
        'id': AiohttpSession(
            server='http://localhost',
            aiohttp_session=sample_session,
        ),
    })
    actual_session = await fn()
    async with actual_session:
        assert actual_session is sample_session


async def test_user_agent_with_a_default_session(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
        ),
    })
    session = await fn()
    async with session:
        assert session.headers['User-Agent'].startswith('kopf/')


async def test_user_agent_with_a_custom_aiohttp_session(vault):
    await vault.populate({
        'id': AiohttpSession(
            server='http://localhost',
            aiohttp_session=aiohttp.ClientSession(),  # no headers, no user-agent
        ),
    })
    session = await fn()
    async with session:
        assert session.headers['User-Agent'].startswith('kopf/')  # injected regardless


async def test_custom_user_agent_preserved(vault):
    headers = {'User-Agent': 'myoperator/1.2.3'}
    await vault.populate({
        'id': AiohttpSession(
            server='http://localhost',
            aiohttp_session=aiohttp.ClientSession(headers=headers),
        ),
    })
    session = await fn()
    async with session:
        assert session.headers['User-Agent'] == 'myoperator/1.2.3'


async def test_trust_env_is_disabled_by_default(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
        ),
    })
    session = await fn()
    async with session:
        assert session.trust_env is False


async def test_trust_env_is_enabled_via_connection_info(vault):
    await vault.populate({
        'id': ConnectionInfo(
            server='http://localhost',
            trust_env=True,
        ),
    })
    session = await fn()
    async with session:
        assert session.trust_env is True


@pytest_asyncio.fixture()
async def proxy():
    server = ProxyServer()
    task = asyncio.create_task(server.run())
    try:
        yield server
    finally:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass  # it is expected (intentionally caused)


@pytest.mark.looptime(noop_cycles=1000)
async def test_proxy(vault, settings, proxy):
    requests = []

    async def callback(request, **_):
        requests.append(request)
        return {'general': 'Kenobi'}

    # Our proxy supports only CONNECT, so only HTTPS. Run ANY https server locally for test.
    # Aresponses runs http. Probing runs http. Webhooks are the only https we have.
    async with WebhookServer() as server:
        async for config in server(callback):
            await vault.populate({
                'id': ConnectionInfo(
                    ca_data=config['caBundle'],
                    server=config['url'],
                    proxy_url=proxy.url,
                ),
            })

            # Check that it was configured client-side (but not necessary used).
            session = await fn()
            assert session._default_proxy == proxy.url

            # Perform the actual request.
            logger = logging.getLogger('irrelevant')
            settings.networking.error_backoffs = []
            resp = await post('/', payload={'hello': 'there'}, logger=logger, settings=settings)

            # Check that the request landed in the destination server, not elsewhere.
            assert requests == [{'hello': 'there'}]
            assert resp == {'general': 'Kenobi'}

            # Check that the request went through the proxy (it was indeed used).
            assert len(proxy.requests) == 1
            proxied_url = f"https://{proxy.requests[0].server_host}:{proxy.requests[0].server_port}"
            assert proxied_url == config['url']

            # Release the resources.
            await session.close()
            break  # webhook-server-specific: we only need one config



import pprint
from typing import Any

import kopf


@kopf.index('pods')
def is_running(*, namespace: str | None, name: str | None, status: kopf.Status, **_: Any) -> Any:
    return {(namespace, name): status.get('phase') == 'Running'}
    # {('kube-system', 'traefik-...-...'): [True],
    #  ('kube-system', 'helm-install-traefik-...'): [False],
    #    ...}


@kopf.index('pods')
def by_label(labels: kopf.Labels, name: str | None, **_: Any) -> Any:
    return {(label, value): name for label, value in labels.items()}
    # {('app', 'traefik'): ['traefik-...-...'],
    #  ('job-name', 'helm-install-traefik'): ['helm-install-traefik-...'],
    #  ('helmcharts.helm.cattle.io/chart', 'traefik'): ['helm-install-traefik-...'],
    #    ...}


@kopf.on.probe()  # type: ignore
def pod_count(is_running: kopf.Index[tuple[str, str], list[bool]], **_: Any) -> int:
    return len(is_running)


@kopf.on.probe()  # type: ignore
def pod_names(is_running: kopf.Index[tuple[str, str], list[bool]], **_: Any) -> list[str]:
    return [name for _, name in is_running]


@kopf.timer('kex', interval=5)  # type: ignore
def intervalled(is_running: kopf.Index[tuple[str, str], list[bool]],
                by_label: kopf.Index[tuple[str, str], list[str]],
                patch: kopf.Patch, **_: Any) -> None:
    pprint.pprint(dict(by_label))
    patch.status['running-pods'] = [
        f"{ns}::{name}"
        for (ns, name), is_running in is_running.items()
        if ns in ['kube-system', 'default']
        if is_running
    ]


# Marks for the e2e tests (see tests/e2e/test_examples.py):
# We do not care: pods can have 6-10 updates here.
E2E_SUCCESS_COUNTS: dict[str, int] = {}



import asyncio

import pytest

import kopf
from kopf._cogs.structs.ephemera import Memo
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import ALL_REASONS
from kopf._core.reactor.inventory import ResourceMemories
from kopf._core.reactor.processing import process_resource_event


@pytest.mark.parametrize('cause_type', ALL_REASONS)
async def test_handlers_called_always(
        registry, settings, handlers, extrahandlers, resource, cause_mock, cause_type,
        assert_logs, k8s_mocked):
    cause_mock.reason = cause_type

    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': 'ev-type', 'object': {'field': 'value'}},
        event_queue=asyncio.Queue(),
    )

    assert handlers.event_mock.call_count == 1
    assert extrahandlers.event_mock.call_count == 1

    event = handlers.event_mock.call_args_list[0].kwargs['event']
    assert 'field' in event['object']
    assert event['object']['field'] == 'value'
    assert event['type'] == 'ev-type'

    assert_logs([
        "Handler 'event_fn' is invoked.",
        "Handler 'event_fn' succeeded.",
        "Handler 'event_fn2' is invoked.",
        "Handler 'event_fn2' succeeded.",
    ])


@pytest.mark.parametrize('cause_type', ALL_REASONS)
async def test_errors_are_ignored(
        registry, settings, handlers, extrahandlers, resource, cause_mock, cause_type,
        assert_logs, k8s_mocked):
    cause_mock.reason = cause_type
    handlers.event_mock.side_effect = Exception("oops")

    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': 'ev-type', 'object': {}},
        event_queue=asyncio.Queue(),
    )

    assert handlers.event_mock.called
    assert extrahandlers.event_mock.called

    assert_logs([
        "Handler 'event_fn' is invoked.",
        "Handler 'event_fn' failed with an exception and will ignore it: oops",
        "Handler 'event_fn2' is invoked.",
        "Handler 'event_fn2' succeeded.",
    ])



"""
Kubernetes watching/streaming and the per-object queueing system.

The framework can handle multiple resources at once.
Every custom resource type is "watched" (as in ``kubectl get --watch``)
in a separate asyncio task in the never-ending loop.

The events for this resource type (of all its objects) are then pushed
to the per-object queues, which are created and destroyed dynamically.
The per-object queues are created on demand.

Every object is identified by its uid, and is handled sequentially:
i.e. the low-level events are processed in the order of their arrival.
Other objects are handled in parallel in their own sequential tasks.

To prevent the memory leaks over the long run, the queues and the workers
of each object are destroyed if no new events arrive for some time.
The destruction delay (usually a few seconds, maybe minutes) is needed
to prevent the often queue/worker destruction and re-creation
in case the events are for any reason delayed by Kubernetes.

The conversion of the low-level watch-events to the high-level causes
is done in the :mod:`kopf._core.reactor.processing` routines.
"""
import asyncio
import contextlib
import enum
import logging
from typing import TYPE_CHECKING, NamedTuple, NewType, Protocol

from kopf._cogs.aiokits import aiotasks, aiotoggles
from kopf._cogs.clients import watching
from kopf._cogs.configs import configuration
from kopf._cogs.structs import bodies, references

logger = logging.getLogger(__name__)


class WatchStreamProcessor(Protocol):
    async def __call__(
            self,
            *,
            raw_event: bodies.RawEvent,
            stream_pressure: asyncio.Event | None = None,  # None for tests
            resource_indexed: aiotoggles.Toggle | None = None,  # None for tests & observation
            operator_indexed: aiotoggles.ToggleSet | None = None,  # None for tests & observation
            consistency_time: float | None = None,  # None for tests
    ) -> str | None:  # patched resource version, if patched
        ...


# An end-of-stream marker sent from the watcher to the workers.
# See: https://www.python.org/dev/peps/pep-0484/#support-for-singleton-types-in-unions
class EOS(enum.Enum):
    token = enum.auto()


if TYPE_CHECKING:
    WatchEventQueue = asyncio.Queue[bodies.RawEvent | EOS]
else:
    WatchEventQueue = asyncio.Queue


class Stream(NamedTuple):
    """ A single object's stream of watch-events, with some extra helpers. """
    backlog: WatchEventQueue
    pressure: asyncio.Event  # means: "hurry up, there are new events queued again"


ObjectUid = NewType('ObjectUid', str)
ObjectRef = tuple[references.Resource, ObjectUid]


def get_uid(raw_event: bodies.RawEvent) -> ObjectUid:
    """
    Retrieve or simulate an identifier of an object unique both in time & space.

    It is used as a key in mappings of framework-internal system resources,
    such as tasks and queues. It is never exposed to the users, even in logs.
    The keys are only persistent during a lifetime of a single process.
    They can be safely changed across different versions.

    In most cases, UIDs are sufficient --- as populated by K8s itself.
    However, some resources have no UIDs: e.g. ``v1/ComponentStatus``:

    .. code-block:: yaml

        apiVersion: v1
        kind: ComponentStatus
        metadata:
          creationTimestamp: null
          name: controller-manager
          selfLink: /api/v1/componentstatuses/controller-manager
        conditions:
        - message: ok
          status: "True"
          type: Healthy

    Note that ``selfLink`` is deprecated and will stop being populated
    since K8s 1.20. Other fields are not always sufficient to ensure uniqueness
    both in space and time: in the example above, the creation time is absent.

    In this function, we do our best to provide a fallback scenario in case
    UIDs are absent. All in all, having slightly less unique identifiers
    is better than failing the whole resource handling completely.
    """
    if 'uid' in raw_event['object']['metadata']:
        uid = raw_event['object']['metadata']['uid']
    else:
        ids = [
            raw_event['object'].get('kind'),
            raw_event['object'].get('apiVersion'),
            raw_event['object']['metadata'].get('name'),
            raw_event['object']['metadata'].get('namespace'),
            raw_event['object']['metadata'].get('creationTimestamp'),
        ]
        uid = '//'.join([s or '-' for s in ids])
    return ObjectUid(uid)


def get_version(raw_event: bodies.RawEvent | EOS) -> str | None:
    if isinstance(raw_event, EOS):
        return None
    else:
        return raw_event.get('object', {}).get('metadata', {}).get('resourceVersion')


async def watcher(
        *,
        namespace: references.Namespace,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        processor: WatchStreamProcessor,
        operator_paused: aiotoggles.ToggleSet | None = None,  # None for tests & observation
        operator_indexed: aiotoggles.ToggleSet | None = None,  # None for tests & observation
        resource_indexed: aiotoggles.Toggle | None = None,  # None for tests & non-indexable
) -> None:
    """
    Watch for the resource events via the API, and spawn the workers per object.

    All resources and objects are done in parallel, but one single object
    is handled sequentially (otherwise, concurrent handling of multiple events
    of the same object could cause data damage).

    The watcher is as non-blocking and async as possible. It does not call any
    external routines, nor does it make the API calls via the sync libraries.

    The watcher is generally a never-ending task (unless an error happens or
    it is cancelled). The workers, on the other hand, are limited approximately
     to the life-time of an object's event.

    Watchers spend their time in the infinite watch stream, not in task waiting.
    The only valid way for a worker to wake up the watcher is to cancel it:
    this will terminate any i/o operation with ``asyncio.CancelledError``, where
    we can make a decision on whether it was a real cancellation, or our own.
    """

    # In case of a failed worker, stop the watcher, and escalate to the operator to stop it.
    watcher_task = asyncio.current_task()
    worker_error: BaseException | None = None
    def exception_handler(exc: BaseException) -> None:
        nonlocal worker_error, watcher_task
        if worker_error is None:
            worker_error = exc
            if watcher_task is not None:  # never happens, but is needed for type-checking.
                watcher_task.cancel()

    # All per-object workers are handled as fire-and-forget jobs via the scheduler,
    # and communicated via the per-object event queues.
    signaller = asyncio.Condition()
    scheduler = aiotasks.Scheduler(limit=settings.queueing.worker_limit,
                                   exception_handler=exception_handler)
    streams: dict[ObjectRef, Stream] = {}

    try:
        # Either use the existing object's queue, or create a new one together with the per-object job.
        # "Fire-and-forget": we do not wait for the result; the job destroys itself when it is fully done.
        stream = watching.infinite_watch(
            settings=settings,
            resource=resource, namespace=namespace,
            operator_paused=operator_paused,
        )
        async for raw_event in stream:

            # If the listing is over (even if it was empty), the resource kind is pre-indexed.
            # At this moment, only the individual workers/processors can block the global readiness.
            if raw_event is watching.Bookmark.LISTED:
                if operator_indexed is not None and resource_indexed is not None:
                    await operator_indexed.drop_toggle(resource_indexed)

            # Whatever is bookmarked there, don't let it go to the multiplexer. Handle it above.
            if isinstance(raw_event, watching.Bookmark):
                continue

            # Skip BOOKMARK events from K8s: they carry no object identity (no uid/name/namespace),
            # and are only useful for resource version tracking (already done in the watch stream).
            if raw_event.get('type') == 'BOOKMARK':
                continue

            # Multiplex the raw events to per-resource workers/queues. Start the new ones if needed.
            key: ObjectRef = (resource, get_uid(raw_event))
            try:
                # Feed the worker, as fast as possible, no extra activities.
                streams[key].pressure.set()  # interrupt current sleeps, if any.
                await streams[key].backlog.put(raw_event)
            except KeyError:

                # Block the operator's readiness for individual resource's index handlers.
                # But NOT when the readiness is already achieved once! After that, ignore it.
                # NB: Strictly before the worker starts -- the processor can be too slow, too late.
                resource_object_indexed: aiotoggles.Toggle | None = None
                if operator_indexed is not None and operator_indexed.is_on():
                    operator_indexed = None
                if operator_indexed is not None and resource_indexed is not None:
                    resource_object_indexed = await operator_indexed.make_toggle(name=f"{key!r}")

                # Start the worker, and feed it initially. Starting can be moderately slow.
                streams[key] = Stream(backlog=asyncio.Queue(), pressure=asyncio.Event())
                streams[key].pressure.set()  # interrupt current sleeps, if any.
                await streams[key].backlog.put(raw_event)
                await scheduler.spawn(
                    name=f'worker for {key}',
                    coro=worker(
                        signaller=signaller,
                        resource_indexed=resource_object_indexed,
                        operator_indexed=operator_indexed,
                        processor=processor,
                        settings=settings,
                        streams=streams,
                        key=key,
                    ))

    except asyncio.CancelledError:
        if worker_error is None:
            raise
        else:
            raise RuntimeError("Event processing has failed with an unrecoverable error. "
                               "This seems to be a framework bug. "
                               "The operator will stop to prevent damage.") from worker_error
    finally:
        # Allow the existing workers to finish gracefully before killing them.
        # Ensure the depletion is done even if the watcher is double-cancelled (e.g. in tests).
        depletion_task = asyncio.create_task(_wait_for_depletion(
            signaller=signaller,
            scheduler=scheduler,
            streams=streams,
            settings=settings,
        ))
        while not depletion_task.done():
            with contextlib.suppress(asyncio.CancelledError):
                await asyncio.shield(depletion_task)

        # Terminate all the fire-and-forget per-object jobs if they are still running.
        # Ensure the scheduler is closed even if the watcher is double-cancelled (e.g. in tests).
        closing_task = asyncio.create_task(scheduler.close())
        while not closing_task.done():
            with contextlib.suppress(asyncio.CancelledError):
                await asyncio.shield(closing_task)


async def worker(
        *,
        signaller: asyncio.Condition,
        settings: configuration.OperatorSettings,
        processor: WatchStreamProcessor,
        resource_indexed: aiotoggles.Toggle | None = None,  # None for tests & observation
        operator_indexed: aiotoggles.ToggleSet | None = None,  # None for tests & observation
        streams: dict[ObjectRef, Stream],
        key: ObjectRef,
) -> None:
    """
    A single worker for a single resource object, each running in its own task.

    An object worker consumes the events from the object-dedicated queue filled
    by the watcher of the whole resource kind (i.e. of all objects of that kind)
    and invokes the processor (which implies handlers) for that specific object.

    The processor is an internal coroutine of the framework, not of the user.
    There are several types of processors involved:

    - For operator peering: monitors the peers and suspends/resumes operations.
    - For cluster observation: spawns new watchers for new CRDs/namespaces.
    - For resource handling: detects changes and calls the user-side handlers.

    The worker is time-limited: it exits as soon as all the object's events
    have been processed and there are no new events for some time of idling
    (a few seconds --- to prevent exiting and recreating the workers too often).
    The watcher will spawn a new worker when (and if) new events arrive.
    Such early exiting saves system resources (RAM) on large clusters with low
    activity, since we do not keep a running worker for every dormant object.
    """
    loop = asyncio.get_running_loop()
    backlog = streams[key].backlog
    pressure = streams[key].pressure
    shouldstop = False
    consistency_time: float | None = None  # None if nothing is expected/awaited.
    expected_version: str | None = None  # None/non-None is synced with the patch-end-time.
    try:
        while not shouldstop:

            # Get an event ASAP (no delay) if possible. But expect the queue can be empty.
            # Save memory by finishing the worker if the backlog is empty for some time.
            timeout = max(settings.queueing.idle_timeout,
                          consistency_time - loop.time() if consistency_time is not None else 0)
            try:
                raw_event = await asyncio.wait_for(backlog.get(), timeout=timeout)
            except asyncio.TimeoutError:
                # A tricky part! Under high-load or with synchronous blocks of asyncio event-loop,
                # it is possible that the timeout happens while the queue is filled: depending on
                # the order in which the coros/waiters are checked once control returns to asyncio.
                # As a work-around, we double-check the queue and exit only if it is truly empty;
                # if not, run as normally. IMPORTANT: There MUST be NO async/await-code between
                # "break" and "finally", so that the queue is not populated again.
                # TODO: LATER: Test the described scenario. I have found no ways to simulate
                #  a timeout while the queue is filled -- neither with pure Python nor with mocks.
                if backlog.empty():
                    break
                else:
                    continue

            # Exit gracefully and immediately on the end-of-stream marker sent by the watcher.
            if isinstance(raw_event, EOS):
                break  # out of the worker.

            # Keep track of the resource's consistency for high-level (state-dependent) handlers.
            # See `settings.persistence.consistency_timeout` for the explanation of consistency.
            if expected_version is not None and expected_version == get_version(raw_event):
                expected_version = None
                consistency_time = None

            # Relieve the pressure only if this is the last event, thus letting the processor sleep.
            # If there are more events to process, fast-skip the sleep as if they have just arrived.
            if backlog.empty():
                pressure.clear()

            # Process the event. It might include sleeping till the time of consistency assumption
            # (i.e. ignoring that the patched version was not received and behaving as if it was).
            newer_patch_version = await processor(
                raw_event=raw_event,
                stream_pressure=pressure,
                resource_indexed=resource_indexed,
                operator_indexed=operator_indexed,
                consistency_time=consistency_time,
            )

            # With every new PATCH API call (if done), restart the consistency waiting.
            if newer_patch_version is not None and settings.persistence.consistency_timeout:
                expected_version = newer_patch_version
                consistency_time = loop.time() + settings.persistence.consistency_timeout

    except Exception:
        # Log the error for every worker: there can be several of them failing at the same time,
        # but only one will trigger the watcher's failure -- others could be lost if not logged.
        logger.exception(f"Event processing has failed with an unrecoverable error for {key}.")
        raise

    finally:
        # Whether an exception or a break or a success, notify the caller, and garbage-collect our queue.
        # The queue must not be left in the queue-cache without a corresponding job handling this queue.
        try:
            del streams[key]
        except KeyError:
            pass  # already absent

        # Notify the depletion routine about the changes in the workers'/streams' overall state.
        # * This should happen STRICTLY AFTER the removal from the streams[], and
        # * This should happen A MOMENT BEFORE the job ends (within the scheduler's close_timeout).
        async with signaller:
            signaller.notify_all()


async def _wait_for_depletion(
        *,
        signaller: asyncio.Condition,
        scheduler: aiotasks.Scheduler,
        settings: configuration.OperatorSettings,
        streams: dict[ObjectRef, Stream],
) -> None:

    # Notify all the workers to finish now. Wake them up if they are waiting in the queue-getting.
    for stream in streams.values():
        await stream.backlog.put(EOS.token)

    # Wait for the queues to be depleted, but only if there are some workers running.
    # Continue with the tasks termination if the timeout is reached, no matter the queues.
    # NB: the scheduler is checked for a case of mocked workers; otherwise, the streams are enough.
    async with signaller:
        try:
            await asyncio.wait_for(
                signaller.wait_for(lambda: not streams or scheduler.empty()),
                timeout=settings.queueing.exit_timeout)
        except asyncio.TimeoutError:
            pass  # if not depleted as configured, proceed with what's left and let it fail

    # The last check if the termination is going to be graceful or not.
    if streams:
        logger.warning(f"Unprocessed streams left for {list(streams.keys())!r}.")



import asyncio

import pytest

from kopf._cogs.aiokits import aiotasks
from kopf._core.reactor.running import run_tasks


async def test_root_task_exit_cancels_remaining(assert_logs, looptime):
    async def finishes_soon():
        await asyncio.sleep(3)

    async def runs_forever():
        await asyncio.sleep(10)

    ignored = asyncio.all_tasks()
    task_a = asyncio.create_task(finishes_soon(), name="finishes_soon")
    task_b = asyncio.create_task(runs_forever(), name="runs_forever")
    await run_tasks([task_a, task_b], ignored=ignored)

    assert task_a.done()
    assert task_b.done()
    assert task_b.cancelled()
    assert looptime == 3
    assert_logs([
        r"Root tasks are stopped: finishing normally",
        r"Hung tasks stopping is skipped: no tasks given",
    ])


async def test_root_task_exception_is_reraised(assert_logs, looptime):
    async def fails_soon():
        await asyncio.sleep(3)
        raise RuntimeError("boom")

    async def runs_forever():
        await asyncio.sleep(10)

    ignored = asyncio.all_tasks()
    task_a = asyncio.create_task(fails_soon(), name="fails_soon")
    task_b = asyncio.create_task(runs_forever(), name="runs_forever")
    with pytest.raises(RuntimeError, match="boom"):
        await run_tasks([task_a, task_b], ignored=ignored)

    assert task_a.done()
    assert task_b.done()
    assert looptime == 3
    assert_logs([
        r"Root tasks are stopped: finishing normally",
    ])


async def test_cancellation_during_root_wait(assert_logs, looptime):
    async def runs_forever():
        await asyncio.sleep(10)

    ignored = asyncio.all_tasks()
    task_a = asyncio.create_task(runs_forever(), name="root_a")
    task_b = asyncio.create_task(runs_forever(), name="root_b")
    runner = asyncio.create_task(run_tasks([task_a, task_b], ignored=ignored))

    await asyncio.sleep(3)
    runner.cancel()

    with pytest.raises(asyncio.CancelledError):
        await runner

    assert task_a.done()
    assert task_b.done()
    assert looptime == 3
    assert_logs([
        r"Root tasks are stopped: cancelling normally",
        r"Hung tasks stopping is skipped: no tasks given",
    ])


async def test_hung_subtask_cancelled_after_grace_period(assert_logs, looptime):
    hung_task: aiotasks.Task | None = None

    async def spawns_and_exits():
        nonlocal hung_task
        hung_task = asyncio.create_task(asyncio.sleep(10), name="hung_subtask")
        await asyncio.sleep(3)

    async def runs_forever():
        await asyncio.sleep(10)

    ignored = asyncio.all_tasks()
    task_a = asyncio.create_task(spawns_and_exits(), name="spawns_and_exits")
    task_b = asyncio.create_task(runs_forever(), name="runs_forever")
    await run_tasks([task_a, task_b], ignored=ignored)

    assert hung_task is not None
    assert hung_task.done()
    assert hung_task.cancelled()
    # 3s for root exit + 5s grace period for hung task timeout.
    assert looptime == 8
    assert_logs([
        r"Root tasks are stopped: finishing normally",
        r"Hung tasks are stopped: finishing normally",
    ])


async def test_hung_subtask_finishes_within_grace_period(assert_logs, looptime):
    hung_task: aiotasks.Task | None = None

    async def spawns_and_exits():
        nonlocal hung_task
        hung_task = asyncio.create_task(asyncio.sleep(2), name="hung_subtask")

    async def runs_forever():
        await asyncio.sleep(10)

    ignored = asyncio.all_tasks()
    task_a = asyncio.create_task(spawns_and_exits(), name="spawns_and_exits")
    task_b = asyncio.create_task(runs_forever(), name="runs_forever")
    await run_tasks([task_a, task_b], ignored=ignored)

    assert hung_task is not None
    assert hung_task.done()
    assert not hung_task.cancelled()
    # 0s for root exit + 2s for hung task finishing within grace period.
    assert looptime == 2
    assert_logs([
        r"Root tasks are stopped: finishing normally",
        r"Hung tasks stopping is skipped: no tasks given",
    ])


async def test_cancellation_during_hung_task_wait(assert_logs, looptime):
    hung_task: aiotasks.Task | None = None

    async def spawns_and_exits():
        nonlocal hung_task
        hung_task = asyncio.create_task(asyncio.sleep(10), name="hung_subtask")
        await asyncio.sleep(3)

    async def runs_forever():
        await asyncio.sleep(10)

    ignored = asyncio.all_tasks()
    task_a = asyncio.create_task(spawns_and_exits(), name="spawns_and_exits")
    task_b = asyncio.create_task(runs_forever(), name="runs_forever")
    runner = asyncio.create_task(run_tasks([task_a, task_b], ignored=ignored))

    # 3s for root exit, then cancel during the 5s hung-task grace period.
    await asyncio.sleep(5)
    runner.cancel()

    with pytest.raises(asyncio.CancelledError):
        await runner

    assert hung_task is not None
    assert hung_task.done()
    assert looptime == 5
    assert_logs([
        r"Root tasks are stopped: finishing normally",
        r"Hung tasks are stopped: cancelling normally",
    ])



import asyncio
import concurrent.futures
import contextlib
import threading
import types
from typing import TYPE_CHECKING, Any, Literal, cast

import click.testing

from kopf import cli
from kopf._cogs.configs import configuration
from kopf._core.intents import registries

_ExcType = BaseException
_ExcInfo = tuple[type[_ExcType], _ExcType, types.TracebackType]

if TYPE_CHECKING:
    ResultFuture = concurrent.futures.Future[click.testing.Result]
    class _AbstractKopfRunner(contextlib.AbstractContextManager["_AbstractKopfRunner"]):
        pass
else:
    ResultFuture = concurrent.futures.Future
    class _AbstractKopfRunner(contextlib.AbstractContextManager):
        pass


class KopfRunner(_AbstractKopfRunner):
    """
    A context manager to run a Kopf-based operator in parallel with the tests.

    Usage:

    .. code-block:: python

        from kopf.testing import KopfRunner

        def test_operator():
            with KopfRunner(['run', '-A', '--verbose', 'examples/01-minimal/example.py']) as runner:
                # do something while the operator is running.
                time.sleep(3)

            assert runner.exit_code == 0
            assert runner.exception is None
            assert 'And here we are!' in runner.output

    All the args & kwargs are passed directly to Click's invocation method.
    See: :class:`click.testing.CliRunner`.
    All properties proxy directly to Click's :class:`click.testing.Result`
    when it is available (i.e. after the context manager exits).

    CLI commands have to be invoked in parallel threads, never in processes:

    First, with multiprocessing, they are unable to pickle and pass
    exceptions (specifically, their traceback objects)
    from a child thread (Kopf's CLI) to the parent thread (pytest).

    Second, mocking works within one process (all threads),
    but not across processes --- the mock's calls (counts, args) are lost.
    """
    _future: ResultFuture

    def __init__(
            self,
            *args: Any,
            reraise: bool = True,
            timeout: float | None = None,
            registry: registries.OperatorRegistry | None = None,
            settings: configuration.OperatorSettings | None = None,
            **kwargs: Any,
    ):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self.reraise = reraise
        self.timeout = timeout
        self.registry = registry
        self.settings = settings
        self._stop = threading.Event()
        self._ready = threading.Event()  # NB: not asyncio.Event!
        self._thread = threading.Thread(target=self._target)
        self._future = concurrent.futures.Future()

    def __enter__(self) -> "KopfRunner":
        self._thread.start()
        self._ready.wait()  # should be nanosecond-fast
        return self

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: types.TracebackType | None,
    ) -> Literal[False]:

        # When the `with` block ends, shut down the parallel thread & loop
        # by cancelling all the tasks. Do not wait for the tasks to finish,
        # but instead wait for the thread+loop (CLI command) to finish.
        self._stop.set()
        self._thread.join(timeout=self.timeout)

        # If the thread is not finished, it is a bigger problem than exceptions.
        if self._thread.is_alive():
            raise Exception("The operator didn't stop, still running.")

        # Re-raise the exceptions of the threading & invocation logic.
        e = self._future.exception()
        if e is not None:
            if exc_val is None:
                raise e
            else:
                raise e from exc_val
        e = self._future.result().exception
        if e is not None and self.reraise:
            if exc_val is None:
                raise e
            else:
                raise e from exc_val

        return False

    def _target(self) -> None:

        # Every thread must have its own loop. The parent thread (pytest)
        # needs to know when the loop is set up, to be able to shut it down.
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self._ready.set()

        # Execute the requested CLI command in the thread & thread's loop.
        # Remember the result & exception for re-raising in the parent thread.
        try:
            ctxobj = cli.CLIControls(
                registry=self.registry,
                settings=self.settings,
                stop_flag=self._stop,
                loop=loop)
            runner = click.testing.CliRunner()
            result = runner.invoke(cli.main, *self.args, **self.kwargs, obj=ctxobj)
        except BaseException as e:
            self._future.set_exception(e)
        else:
            self._future.set_result(result)
        finally:

            # Shut down the API-watching streams.
            loop.run_until_complete(loop.shutdown_asyncgens())

            # Shut down the transports and prevent ResourceWarning: unclosed transport.
            # See: https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
            # TODO: Try a hack: https://github.com/aio-libs/aiohttp/issues/1925#issuecomment-575754386
            loop.run_until_complete(asyncio.sleep(1.0))

            loop.close()

    @property
    def future(self) -> ResultFuture:
        return self._future

    @property
    def output(self) -> str:
        return self.future.result().output

    @property
    def stdout(self) -> str:
        return self.future.result().stdout

    @property
    def stdout_bytes(self) -> bytes:
        return self.future.result().stdout_bytes

    @property
    def stderr(self) -> str:
        return self.future.result().stderr

    @property
    def stderr_bytes(self) -> bytes:
        return self.future.result().stderr_bytes or b''

    @property
    def exit_code(self) -> int:
        return self.future.result().exit_code

    @property
    def exception(self) -> _ExcType:
        return cast(_ExcType, self.future.result().exception)

    @property
    def exc_info(self) -> _ExcInfo:
        return cast(_ExcInfo, self.future.result().exc_info)



import types

import pytest

from kopf._cogs.helpers.thirdparty import KubernetesModelSync


@pytest.mark.parametrize('name', ['V1Pod', 'V1ObjectMeta', 'V1PodSpec', 'V1PodTemplateSpec'])
def test_kubernetes_model_classes_detection(kubernetes, name):
    cls = getattr(kubernetes.client, name)
    assert issubclass(cls, KubernetesModelSync)


@pytest.mark.parametrize('name', ['CoreV1Api', 'ApiClient', 'Configuration'])
def test_kubernetes_other_classes_detection(kubernetes, name):
    cls = getattr(kubernetes.client, name)
    assert not issubclass(cls, KubernetesModelSync)


@pytest.mark.parametrize('cls', [object, types.SimpleNamespace])
def test_non_kubernetes_classes_detection(kubernetes, cls):
    assert not issubclass(cls, KubernetesModelSync)



import pytest

from kopf._cogs.structs.diffs import Diff, DiffItem, DiffOperation, reduce

DIFF = Diff([
    DiffItem(DiffOperation.ADD   , ('key1',), None, 'new1'),
    DiffItem(DiffOperation.CHANGE, ('key2',), 'old2', 'new2'),
    DiffItem(DiffOperation.ADD   , ('key2', 'suba'), 'olda', 'newa'),
    DiffItem(DiffOperation.REMOVE, ('key2', 'subb'), 'oldb', 'newb'),
    DiffItem(DiffOperation.REMOVE, ('key3',), 'old3', None),
    DiffItem(DiffOperation.CHANGE, ('key4',),
             {'suba': 'olda', 'subc': 'oldc'},
             {'subb': 'newb', 'subc': 'newc'}),
])


@pytest.mark.parametrize('diff', [
    [['op', ['key', 'sub'], 'old', 'new']],
    [['op', ('key', 'sub'), 'old', 'new']],
    [('op', ['key', 'sub'], 'old', 'new')],
    [('op', ('key', 'sub'), 'old', 'new')],
    (['op', ['key', 'sub'], 'old', 'new'],),
    (['op', ('key', 'sub'), 'old', 'new'],),
    (('op', ['key', 'sub'], 'old', 'new'),),
    (('op', ('key', 'sub'), 'old', 'new'),),
], ids=[
    'lll-diff', 'llt-diff', 'ltl-diff', 'ltt-diff',
    'tll-diff', 'tlt-diff', 'ttl-diff', 'ttt-diff',
])
@pytest.mark.parametrize('path', [
    ['key', 'sub'],
    ('key', 'sub'),
], ids=['list-path', 'tuple-path'])
def test_type_ignored_for_inputs_but_is_tuple_for_output(diff, path):
    result = reduce(diff, path)
    assert result == (('op', (), 'old', 'new'),)


def test_empty_path_selects_all_ops():
    result = reduce(DIFF, [])
    assert result == DIFF


def test_existent_path_selects_relevant_ops():
    result = reduce(DIFF, ['key2'])
    assert result == (
        ('change', (), 'old2', 'new2'),
        ('add'   , ('suba',), 'olda', 'newa'),
        ('remove', ('subb',), 'oldb', 'newb'),
    )


@pytest.mark.parametrize('path', [
    ['nonexistent-key'],
    ['key1', 'nonexistent-key'],
    ['key2', 'nonexistent-key'],
    ['key3', 'nonexistent-key'],
    ['key4', 'nonexistent-key'],
    ['key4', 'suba', 'nonexistent-key'],
    ['key4', 'subb', 'nonexistent-key'],
    ['key4', 'subc', 'nonexistent-key'],
    ['key4', 'nonexistent-dict', 'nonexistent-key'],
])
def test_nonexistent_path_selects_nothing(path):
    result = reduce(DIFF, path)
    assert result == ()


def test_overly_specific_path_dives_into_dicts_for_addition():
    result = reduce(DIFF, ['key4', 'subb'])
    assert result == (
        ('add', (), None, 'newb'),
    )


def test_overly_specific_path_dives_into_dicts_for_removal():
    result = reduce(DIFF, ['key4', 'suba'])
    assert result == (
        ('remove', (), 'olda', None),
    )


def test_overly_specific_path_dives_into_dicts_for_change():
    result = reduce(DIFF, ['key4', 'subc'])
    assert result == (
        ('change', (), 'oldc', 'newc'),
    )



import logging

import pytest

import kopf


async def test_declared_public_interface_and_promised_defaults():
    settings = kopf.OperatorSettings()
    assert settings.posting.level == logging.INFO
    assert settings.peering.name == "default"
    assert settings.peering.stealth == False
    assert settings.peering.priority == 0
    assert settings.peering.lifetime == 60
    assert settings.peering.mandatory == False
    assert settings.peering.standalone == False
    assert settings.peering.namespaced == True
    assert settings.peering.clusterwide == False
    assert settings.watching.reconnect_backoff == 0.1
    assert settings.watching.connect_timeout is None
    assert settings.watching.server_timeout is None
    assert settings.watching.client_timeout is None
    assert settings.queueing.worker_limit is None
    assert settings.queueing.idle_timeout == 5.0
    assert settings.queueing.exit_timeout == 2.0
    assert settings.queueing.error_delays == (1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610)
    assert settings.scanning.disabled == False
    assert settings.admission.server is None
    assert settings.admission.managed is None
    assert settings.execution.executor is not None
    assert settings.execution.max_workers is None
    assert settings.networking.request_timeout == 5 * 60
    assert settings.networking.connect_timeout is None
    assert settings.networking.trust_env == False
    assert settings.persistence.consistency_timeout == 5.0


async def test_peering_namespaced_is_modified_by_clusterwide():
    settings = kopf.OperatorSettings()
    assert settings.peering.namespaced == True
    settings.peering.clusterwide = not settings.peering.clusterwide
    assert settings.peering.namespaced == False


async def test_peering_clusterwide_is_modified_by_namespaced():
    settings = kopf.OperatorSettings()
    assert settings.peering.clusterwide == False
    settings.peering.namespaced = not settings.peering.namespaced
    assert settings.peering.clusterwide == True


async def test_deprecated_batching_settings():
    settings = kopf.OperatorSettings()
    with pytest.warns(DeprecationWarning, match=r"Batching settings are now queueing settings."):
        assert settings.batching.worker_limit is None


async def test_deprecated_batch_window_is_still_persisted():
    settings = kopf.OperatorSettings()
    with pytest.warns(DeprecationWarning, match=r"Time-based event batching was removed."):
        assert settings.queueing.batch_window == 0.1
    with pytest.warns(DeprecationWarning, match=r"Time-based event batching was removed."):
        settings.queueing.batch_window = 123
    with pytest.warns(DeprecationWarning, match=r"Time-based event batching was removed."):
        assert settings.queueing.batch_window == 123  # persisted as set above



"""
Verify that the module-prohibiting fixtures do work as expected.
Otherwise, the tests are useless or can show false-positives.
"""
import pytest


@pytest.mark.usefixtures('no_kubernetes_asyncio')
def test_client_asyncio_uninstalled_has_effect():
    with pytest.raises(ImportError):
        import kubernetes_asyncio


@pytest.mark.usefixtures('no_kubernetes')
def test_client_uninstalled_has_effect():
    with pytest.raises(ImportError):
        import kubernetes


@pytest.mark.usefixtures('no_pykube')
def test_pykube_uninstalled_has_effect():
    with pytest.raises(ImportError):
        import pykube



import dataclasses
import warnings
from collections.abc import Collection

from kopf._cogs.structs import dicts, diffs, references
from kopf._core.actions import execution
from kopf._core.intents import callbacks, causes, filters


@dataclasses.dataclass(frozen=True)
class ActivityHandler(execution.Handler):
    fn: callbacks.ActivityFn  # typing clarification
    activity: causes.Activity | None
    _fallback: bool = False  # non-public!

    def __str__(self) -> str:
        return f"Activity {self.id!r}"


@dataclasses.dataclass(frozen=True)
class ResourceHandler(execution.Handler):
    selector: references.Selector | None  # None is used only in sub-handlers
    labels: filters.MetaFilter | None
    annotations: filters.MetaFilter | None
    when: callbacks.WhenFilterFn | None
    field: dicts.FieldPath | None
    value: filters.ValueFilter | None

    def adjust_cause(self, cause: execution.CauseT) -> execution.CauseT:
        if self.field is not None and isinstance(cause, causes.ChangingCause):
            old = dicts.resolve(cause.old, self.field, None)
            new = dicts.resolve(cause.new, self.field, None)
            diff = diffs.reduce(cause.diff, self.field)
            return dataclasses.replace(cause, old=old, new=new, diff=diff)
        else:
            return cause


@dataclasses.dataclass(frozen=True)
class WebhookHandler(ResourceHandler):
    fn: callbacks.WebhookFn  # typing clarification
    reason: causes.WebhookType
    operations: Collection[str] | None
    subresource: str | None
    persistent: bool | None
    side_effects: bool | None
    ignore_failures: bool | None

    def __str__(self) -> str:
        return f"Webhook {self.id!r}"

    @property
    def operation(self) -> str | None:  # deprecated
        warnings.warn("handler.operation is deprecated, use handler.operations", DeprecationWarning)
        if not self.operations:
           return None
        elif len(self.operations) == 1:
            return list(self.operations)[0]
        else:
            raise ValueError(
                f"{len(self.operations)} operations in the handler. Use it as handler.operations."
            )


@dataclasses.dataclass(frozen=True)
class IndexingHandler(ResourceHandler):
    fn: callbacks.IndexingFn  # typing clarification

    def __str__(self) -> str:
        return f"Indexer {self.id!r}"


@dataclasses.dataclass(frozen=True)
class WatchingHandler(ResourceHandler):
    fn: callbacks.WatchingFn  # typing clarification


@dataclasses.dataclass(frozen=True)
class ChangingHandler(ResourceHandler):
    fn: callbacks.ChangingFn  # typing clarification
    reason: causes.Reason | None
    initial: bool | None
    deleted: bool | None  # used for mixed-in (initial==True) @on.resume handlers only.
    requires_finalizer: bool | None
    field_needs_change: bool | None  # to identify on-field/on-update with support for old=/new=.
    old: filters.ValueFilter | None
    new: filters.ValueFilter | None


@dataclasses.dataclass(frozen=True)
class SpawningHandler(ResourceHandler):
    requires_finalizer: bool | None
    initial_delay: float | callbacks.DelayFn | None


@dataclasses.dataclass(frozen=True)
class DaemonHandler(SpawningHandler):
    fn: callbacks.DaemonFn  # typing clarification
    cancellation_backoff: float | None  # how long to wait before actual cancellation.
    cancellation_timeout: float | None  # how long to wait before giving up on cancellation.
    cancellation_polling: float | None  # how often to check for cancellation status.

    def __str__(self) -> str:
        return f"Daemon {self.id!r}"


@dataclasses.dataclass(frozen=True)
class TimerHandler(SpawningHandler):
    fn: callbacks.TimerFn  # typing clarification
    sharp: bool | None
    idle: float | None
    interval: float | None

    def __str__(self) -> str:
        return f"Timer {self.id!r}"



import asyncio
import contextlib
import signal

from kopf._core.reactor.running import stop_flag_checker


async def test_stop_via_stop_flag_with_none_result(assert_logs, looptime):
    signal_flag: asyncio.Future[object] = asyncio.Future()
    stop_flag: asyncio.Future[object] = asyncio.Future()

    task = asyncio.create_task(stop_flag_checker(
        signal_flag=signal_flag,
        stop_flag=stop_flag,
    ))

    await asyncio.sleep(3)
    stop_flag.set_result(None)
    await task

    assert_logs(["Stop-flag is raised. Operator is stopping."])
    assert looptime == 3


async def test_stop_via_stop_flag_with_custom_value(assert_logs, looptime):
    signal_flag: asyncio.Future[object] = asyncio.Future()
    stop_flag: asyncio.Future[object] = asyncio.Future()

    task = asyncio.create_task(stop_flag_checker(
        signal_flag=signal_flag,
        stop_flag=stop_flag,
    ))

    await asyncio.sleep(3)
    stop_flag.set_result('custom-reason')
    await task

    assert_logs(["Stop-flag is set to 'custom-reason'. Operator is stopping."])
    assert looptime == 3


async def test_stop_via_asyncio_event(assert_logs, looptime):
    signal_flag: asyncio.Future[object] = asyncio.Future()
    stop_flag = asyncio.Event()

    task = asyncio.create_task(stop_flag_checker(
        signal_flag=signal_flag,
        stop_flag=stop_flag,
    ))

    await asyncio.sleep(3)
    stop_flag.set()
    await task

    assert_logs(["Stop-flag is set to True. Operator is stopping."])
    assert looptime == 3


async def test_stop_via_signal_flag(assert_logs, looptime):
    signal_flag: asyncio.Future[object] = asyncio.Future()

    task = asyncio.create_task(stop_flag_checker(
        signal_flag=signal_flag,
        stop_flag=None,
    ))

    await asyncio.sleep(3)
    signal_flag.set_result(signal.SIGINT)
    await task

    assert_logs(["Signal SIGINT is received. Operator is stopping."])
    assert looptime == 3


async def test_cancellation_exits_silently(assert_logs, looptime):
    signal_flag: asyncio.Future[object] = asyncio.Future()

    task = asyncio.create_task(stop_flag_checker(
        signal_flag=signal_flag,
        stop_flag=None,
    ))

    await asyncio.sleep(3)
    task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await task

    assert_logs(prohibited=["Operator is stopping"])
    assert looptime == 3



import functools

import pytest

from kopf._cogs.structs.finalizers import allow_deletion, block_deletion, \
                                          is_deletion_blocked, is_deletion_ongoing
from kopf._cogs.structs.patches import Patch


def test_finalizer_is_fqdn(settings):
    assert settings.persistence.finalizer.startswith('kopf.zalando.org/')


@pytest.mark.parametrize('expected, body', [
    pytest.param(True, {'metadata': {'deletionTimestamp': '2020-12-31T23:59:59'}}, id='time'),
    pytest.param(False, {'metadata': {'deletionTimestamp': None}}, id='none'),
    pytest.param(False, {'metadata': {}}, id='no-field'),
    pytest.param(False, {}, id='no-metadata'),
])
def test_is_deleted(expected, body):
    result = is_deletion_ongoing(body=body)
    assert result == expected


@pytest.mark.parametrize('expected, body', [
    pytest.param(False, {}, id='no-metadata'),
    pytest.param(False, {'metadata': {}}, id='no-finalizers'),
    pytest.param(False, {'metadata': {'finalizers': []}}, id='empty'),
    pytest.param(False, {'metadata': {'finalizers': ['other']}}, id='others'),
    pytest.param(True, {'metadata': {'finalizers': ['fin']}}, id='normal'),
    pytest.param(True, {'metadata': {'finalizers': ['other', 'fin']}}, id='mixed'),
])
def test_has_finalizers(expected, body):
    result = is_deletion_blocked(body=body, finalizer='fin')
    assert result == expected


def test_append_finalizers_to_others():
    body = {'metadata': {'finalizers': ['other1', 'other2'], 'resourceVersion': '1234567890'}}
    patch = Patch(fns=[functools.partial(block_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == [{'op': 'add', 'path': '/metadata/finalizers/2', 'value': 'fin'}]


def test_append_finalizers_to_empty():
    body = {}
    patch = Patch(fns=[functools.partial(block_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == [{'op': 'add', 'path': '/metadata', 'value': {'finalizers': ['fin']}}]


def test_append_finalizers_when_present():
    body = {'metadata': {'finalizers': ['other1', 'fin', 'other2']}}
    patch = Patch(fns=[functools.partial(block_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == []


def test_remove_finalizers_keeps_others():
    body = {'metadata': {'finalizers': ['other1', 'fin', 'other2'], 'resourceVersion': '1234567890'}}
    patch = Patch(fns=[functools.partial(allow_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == [{'op': 'remove', 'path': '/metadata/finalizers/1'}]


def test_remove_finalizers_cleans_keys():
    body = {'metadata': {'finalizers': ['fin'], 'resourceVersion': '1234567890'}}
    patch = Patch(fns=[functools.partial(allow_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == [{'op': 'remove', 'path': '/metadata/finalizers'}]


def test_remove_finalizers_cleans_metadata():
    body = {'metadata': {'finalizers': ['fin']}}
    patch = Patch(fns=[functools.partial(allow_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == [{'op': 'remove', 'path': '/metadata'}]


def test_remove_finalizers_when_absent():
    body = {'metadata': {'finalizers': ['other1', 'other2']}}
    patch = Patch(fns=[functools.partial(allow_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == []


def test_remove_finalizers_when_empty():
    body = {}
    patch = Patch(fns=[functools.partial(allow_deletion, finalizer='fin')])
    ops = patch.as_json_patch(body)
    assert ops == []



import asyncio
import contextlib

import pytest

import kopf
from kopf._cogs.aiokits import aiotasks
from kopf._cogs.structs.credentials import Vault
from kopf._cogs.structs.ephemera import AnyMemo, Memo
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.reactor.running import startup_cleanup_activities


async def test_startup_sets_flags(registry, settings, looptime):
    started_flag = asyncio.Event()
    ready_flag = asyncio.Event()
    root_tasks: list[aiotasks.Task] = []
    core_task = asyncio.create_task(asyncio.sleep(10), name="dummy_core")

    task = asyncio.create_task(startup_cleanup_activities(
        root_tasks=root_tasks,
        core_tasks=[core_task],
        ready_flag=ready_flag,
        started_flag=started_flag,
        registry=registry,
        settings=settings,
        indices=OperatorIndexers().indices,
        vault=Vault(),
        memo=AnyMemo(Memo()),
    ))
    root_tasks.append(task)

    await started_flag.wait()
    assert ready_flag.is_set()

    task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await task

    assert core_task.done()
    assert looptime == 0


async def test_cleanup_runs_only_after_root_tasks_exit(registry, settings, looptime):
    started_flag = asyncio.Event()
    root_tasks: list[aiotasks.Task] = []
    cleanup_result: dict[str, object] = {}
    core_task = asyncio.create_task(asyncio.sleep(10), name="dummy_core")

    @kopf.on.cleanup(registry=registry)
    async def cleanup_handler(**_):
        cleanup_result['called'] = True

    async def dummy_task():
        await asyncio.sleep(3)

    dummy = asyncio.create_task(dummy_task(), name="dummy_task")
    root_tasks.append(dummy)

    task = asyncio.create_task(startup_cleanup_activities(
        root_tasks=root_tasks,
        core_tasks=[core_task],
        ready_flag=None,
        started_flag=started_flag,
        registry=registry,
        settings=settings,
        indices=OperatorIndexers().indices,
        vault=Vault(),
        memo=AnyMemo(Memo()),
    ))
    root_tasks.append(task)

    await started_flag.wait()

    # Simulate operator shutdown: cancel the activities task (wakes it from sleep).
    task.cancel()
    await asyncio.sleep(0)

    # The activities task is now waiting for other root tasks to exit.
    # The dummy task sleeps for 3 seconds, then exits.
    with contextlib.suppress(asyncio.CancelledError):
        await task

    assert core_task.done()
    assert cleanup_result == {'called': True}
    assert looptime == 3


async def test_cancellation_during_startup(registry, settings, assert_logs, looptime):
    started_flag = asyncio.Event()
    core_task = asyncio.create_task(asyncio.sleep(10), name="dummy_core")

    @kopf.on.startup(registry=registry)
    async def blocking_startup(**_):
        await asyncio.sleep(10)

    task = asyncio.create_task(startup_cleanup_activities(
        root_tasks=[],
        core_tasks=[core_task],
        ready_flag=None,
        started_flag=started_flag,
        registry=registry,
        settings=settings,
        indices=OperatorIndexers().indices,
        vault=Vault(),
        memo=AnyMemo(Memo()),
    ))

    await asyncio.sleep(3)

    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task

    assert core_task.done()
    assert not started_flag.is_set()
    assert_logs(["Startup activity is only partially executed due to cancellation."])
    assert looptime == 3


async def test_cancellation_during_root_task_wait(registry, settings, assert_logs, looptime):
    started_flag = asyncio.Event()
    root_tasks: list[aiotasks.Task] = []
    core_task = asyncio.create_task(asyncio.sleep(10), name="dummy_core")

    async def dummy_task():
        await asyncio.sleep(10)

    dummy = asyncio.create_task(dummy_task(), name="dummy_task")
    root_tasks.append(dummy)

    task = asyncio.create_task(startup_cleanup_activities(
        root_tasks=root_tasks,
        core_tasks=[core_task],
        ready_flag=None,
        started_flag=started_flag,
        registry=registry,
        settings=settings,
        indices=OperatorIndexers().indices,
        vault=Vault(),
        memo=AnyMemo(Memo()),
    ))
    root_tasks.append(task)

    await started_flag.wait()

    # First cancel wakes from sleep; the task then waits for root tasks.
    task.cancel()
    await asyncio.sleep(0)

    # Wait a bit, then cancel again while it is waiting for the dummy task.
    await asyncio.sleep(3)
    task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await task

    assert core_task.done()
    assert_logs(["Cleanup activity is not executed at all due to cancellation."])
    assert looptime == 3

    dummy.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await dummy


async def test_cancellation_during_cleanup(registry, settings, assert_logs, looptime):
    started_flag = asyncio.Event()
    root_tasks: list[aiotasks.Task] = []
    core_task = asyncio.create_task(asyncio.sleep(10), name="dummy_core")

    @kopf.on.cleanup(registry=registry)
    async def blocking_cleanup(**_):
        await asyncio.sleep(10)

    async def dummy_task():
        await asyncio.sleep(3)

    dummy = asyncio.create_task(dummy_task(), name="dummy_task")
    root_tasks.append(dummy)

    task = asyncio.create_task(startup_cleanup_activities(
        root_tasks=root_tasks,
        core_tasks=[core_task],
        ready_flag=None,
        started_flag=started_flag,
        registry=registry,
        settings=settings,
        indices=OperatorIndexers().indices,
        vault=Vault(),
        memo=AnyMemo(Memo()),
    ))
    root_tasks.append(task)

    await started_flag.wait()

    # First cancel wakes from sleep; the task then waits for root tasks.
    task.cancel()
    await asyncio.sleep(0)

    # The dummy task finishes after 3 seconds, so the cleanup handler starts.
    # Wait a bit more, then cancel during the cleanup handler.
    await asyncio.sleep(6)
    task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await task

    assert core_task.done()
    assert_logs(["Cleanup activity is only partially executed due to cancellation."])
    assert looptime == 6


async def test_cancellation_during_core_task_stop(registry, settings, assert_logs, looptime):
    started_flag = asyncio.Event()
    root_tasks: list[aiotasks.Task] = []

    async def dummy_root_task():
        await asyncio.sleep(3)

    async def stubborn_core_task():
        try:
            await asyncio.sleep(10)
        except asyncio.CancelledError:
            await asyncio.sleep(10)

    dummy = asyncio.create_task(dummy_root_task(), name="dummy_root")
    root_tasks.append(dummy)

    core_tasks = [asyncio.create_task(stubborn_core_task(), name="stubborn_core")]

    task = asyncio.create_task(startup_cleanup_activities(
        root_tasks=root_tasks,
        core_tasks=core_tasks,
        ready_flag=None,
        started_flag=started_flag,
        registry=registry,
        settings=settings,
        indices=OperatorIndexers().indices,
        vault=Vault(),
        memo=AnyMemo(Memo()),
    ))
    root_tasks.append(task)

    await started_flag.wait()

    # First cancel wakes from sleep; the task then waits for root tasks.
    task.cancel()
    await asyncio.sleep(0)

    # The root task finishes after 3 seconds, then core task stopping begins.
    # Wait a bit more, then cancel during the core task stopping phase.
    await asyncio.sleep(5)
    task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await task

    assert looptime == 5
    assert_logs(["Cleanup activity is not executed at all due to cancellation."])

    # Clean up the stubborn core task.
    core_tasks[0].cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await core_tasks[0]


async def test_core_tasks_are_stopped_after_root_tasks(settings, registry, looptime):
    timestamps: dict[str, float] = {}

    async def sample_task(duration: float, key: str) -> None:
        nonlocal timestamps
        try:
            await asyncio.sleep(duration)
        finally:
            timestamps[key] = float(looptime)

    root_tasks = [
        asyncio.create_task(sample_task(2, 'root1')),
        asyncio.create_task(sample_task(5, 'root2')),
    ]
    core_tasks = [
        asyncio.create_task(sample_task(9, 'core1')),
        asyncio.create_task(sample_task(9, 'core2')),
    ]

    started_flag = asyncio.Event()
    activities_task = asyncio.create_task(startup_cleanup_activities(
        root_tasks=root_tasks,
        core_tasks=core_tasks,
        ready_flag=None,
        started_flag=started_flag,
        registry=registry,
        settings=settings,
        indices=OperatorIndexers().indices,
        vault=Vault(),
        memo=Memo(),
    ))
    root_tasks.append(activities_task)

    # Simulate the behavior of the guarded task (briefly).
    await started_flag.wait()
    activities_task.cancel()  # interrupt its eternal sleep
    await activities_task

    assert timestamps['root1'] == 2
    assert timestamps['root2'] == 5
    assert timestamps['core1'] == 5  # the latest of root tasks, not 9
    assert timestamps['core2'] == 5  # the latest of root tasks, not 9



import asyncio
import json

import pytest

import kopf
from kopf._cogs.structs.ephemera import Memo
from kopf._core.actions.execution import PermanentError, TemporaryError
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import HANDLER_REASONS, Reason
from kopf._core.reactor.inventory import ResourceMemories
from kopf._core.reactor.processing import process_resource_event


# The extrahandlers are needed to prevent the cycle ending and status purging.
@pytest.mark.parametrize('cause_type', HANDLER_REASONS)
async def test_fatal_error_stops_handler(
        registry, settings, handlers, extrahandlers, resource, cause_mock, cause_type,
        assert_logs, k8s_mocked, looptime):
    name1 = f'{cause_type}_fn'

    event_type = None if cause_type == Reason.RESUME else 'irrelevant'
    cause_mock.reason = cause_type
    handlers.create_mock.side_effect = PermanentError("oops")
    handlers.update_mock.side_effect = PermanentError("oops")
    handlers.delete_mock.side_effect = PermanentError("oops")
    handlers.resume_mock.side_effect = PermanentError("oops")

    await process_resource_event(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=asyncio.Queue(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == Reason.RESUME else 0)

    assert looptime == 0
    assert k8s_mocked.patch.called

    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert patch['metadata']['annotations']  # not empty at least
    progress = json.loads(patch['metadata']['annotations'][f"kopf.zalando.org/{name1}"])
    assert progress['failure'] is True
    assert progress['message'] == 'oops'

    assert_logs([
        "Handler .+ failed permanently: oops",
    ])


# The extrahandlers are needed to prevent the cycle ending and status purging.
@pytest.mark.parametrize('cause_type', HANDLER_REASONS)
async def test_retry_error_delays_handler(
        registry, settings, handlers, extrahandlers, resource, cause_mock, cause_type,
        assert_logs, k8s_mocked, looptime):
    name1 = f'{cause_type}_fn'

    event_type = None if cause_type == Reason.RESUME else 'irrelevant'
    cause_mock.reason = cause_type
    handlers.create_mock.side_effect = TemporaryError("oops")
    handlers.update_mock.side_effect = TemporaryError("oops")
    handlers.delete_mock.side_effect = TemporaryError("oops")
    handlers.resume_mock.side_effect = TemporaryError("oops")

    await process_resource_event(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=asyncio.Queue(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == Reason.RESUME else 0)

    assert looptime == 0
    assert k8s_mocked.patch.called

    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert patch['metadata']['annotations']  # not empty at least
    progress = json.loads(patch['metadata']['annotations'][f"kopf.zalando.org/{name1}"])
    assert progress['failure'] is False
    assert progress['success'] is False
    assert progress['delayed']

    assert_logs([
        "Handler .+ failed temporarily: oops",
    ])


# The extrahandlers are needed to prevent the cycle ending and status purging.
@pytest.mark.parametrize('cause_type', HANDLER_REASONS)
async def test_arbitrary_error_delays_handler(
        registry, settings, handlers, extrahandlers, resource, cause_mock, cause_type,
        assert_logs, k8s_mocked, looptime):
    name1 = f'{cause_type}_fn'

    event_type = None if cause_type == Reason.RESUME else 'irrelevant'
    cause_mock.reason = cause_type
    handlers.create_mock.side_effect = Exception("oops")
    handlers.update_mock.side_effect = Exception("oops")
    handlers.delete_mock.side_effect = Exception("oops")
    handlers.resume_mock.side_effect = Exception("oops")

    await process_resource_event(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=asyncio.Queue(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == Reason.RESUME else 0)

    assert looptime == 0
    assert k8s_mocked.patch.called

    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert patch['metadata']['annotations']  # not empty at least
    progress = json.loads(patch['metadata']['annotations'][f"kopf.zalando.org/{name1}"])
    assert progress['failure'] is False
    assert progress['success'] is False
    assert progress['delayed']

    assert_logs([
        "Handler .+ failed with an exception and will try again in 60 seconds: oops",
    ])
